# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""A time tracker for the command line. Utilizing the power of hamster! [nark]."""

# BREADCRUMB: PROFILING
from nark.helpers.dev.profiling import profile_elapsed
# BREADCRUMB: PROFILING
# E402 module level import not at top of file
profile_elapsed('To dob:   _top')  # noqa: E402

from gettext import gettext as _

import os
import re
import sys
from functools import update_wrapper

import click_hotoffthehamster as click

from dob_bright.crud.fact_dressed import FactDressed
from dob_bright.styling.ignore_cmds import (
    create_ignore_conf,
    echo_ignore_sections,
    echo_ignore_table,
    edit_ignore_file
)
from dob_bright.styling.rules_cmds import (
    create_rules_conf,
    echo_rules_conf,
    echo_rule_names,
    echo_rules_table,
    edit_rules_conf
)
from dob_bright.styling.styles_cmds import (
    create_styles_conf,
    echo_styles_conf,
    echo_styles_list,
    echo_styles_table,
    edit_styles_conf
)
from dob_bright.termio.echoes import click_echo
from dob_bright.termio.errors import dob_in_user_exit
from dob_bright.termio.paging import flush_pager

from .clickux import help_strings
from .clickux import help_string_add_fact
from .clickux.add_fact_help_group import ClickAddFactHelpGroup
from .clickux.aliasable_bunchy_plugin import ClickAliasableBunchyPluginGroup
from .clickux.bunchy_help import (
    cmd_bunch_group_introducing,
    cmd_bunch_group_edit,
    cmd_bunch_group_get_meta,
    cmd_bunch_group_generate_report,
    cmd_bunch_group_dbms,
    cmd_bunch_group_add_fact,
    cmd_bunch_group_ongoing_fact,
    cmd_bunch_group_personalize
)
from .clickux.cmd_options import (
    cmd_options_edit_item,
    cmd_options_fact_add,
    cmd_options_fact_dryable,
    cmd_options_fact_edit,
    cmd_options_fact_import,
    cmd_options_factoid,
    cmd_options_factoid_verify_none,
    cmd_options_factoid_verify_start,
    cmd_options_factoid_verify_end,
    cmd_options_factoid_verify_both,
    cmd_options_rule_name,
    cmd_options_styles_internal,
    cmd_options_styles_named
)
from .clickux.cmd_options_search import (
    cmd_options_any_search_query,
    cmd_options_output_format_any_input,
    postprocess_options_normalize_search_args,
    postprocess_options_output_format_any_input
)
from .clickux.help_command import help_command_help
from .clickux.help_detect import show_help_finally, show_help_if_no_command
from .clickux.induct_newbies import induct_newbies, insist_germinated
from .clickux.plugin_group import ensure_plugged_in
from .clickux.post_processor import post_processor
from .cmds_list import activity as list_activity
from .cmds_list import category as list_category
from .cmds_list import fact as list_fact
from .cmds_list import tag as list_tag
from .cmds_usage import activity as usage_activity
from .cmds_usage import category as usage_category
from .cmds_usage import tag as usage_tag
from .complete import tab_complete
from .config import (
    echo_config_table,
    echo_config_value,
    edit_config_file,
    write_config_value
)
from .copyright import echo_copyright, echo_license
from .demo import demo_config, demo_dob
from .details import echo_app_details, echo_app_environs, echo_data_stats
from .facts.add_fact import add_fact
from .facts.cancel_fact import cancel_fact
from .facts.echo_fact import echo_latest_ended, echo_ongoing_fact, echo_ongoing_or_ended
from .facts.edit_fact import edit_fact_by_pk
from .facts.import_facts import import_facts
from .migrate import control as migrate_control
from .migrate import downgrade as migrate_downgrade
from .migrate import upgrade as migrate_upgrade
from .migrate import upgrade_legacy_database_file
from .migrate import version as migrate_version
from .run_cli import dob_versions, pass_controller, pass_controller_context, run

# __all__ = ( ... )  # So many. Too tedious to list.


# ***
# *** [HELP]
# ***

@cmd_bunch_group_get_meta
@run.command(help=help_strings.HELP_HELP)
# (lb): Should not need the help-finally decorator, but doesn't seem to hurt, either:
#   @show_help_finally
@flush_pager
@click.pass_context
@click.argument('command', nargs=-1)
def help(ctx, command=None):
    """Show help."""
    help_command_help(ctx, command)


# ***
# *** [VERSION] Ye rote version command.
# ***

@cmd_bunch_group_get_meta
@run.command(help=help_strings.VERSION_HELP)
@show_help_finally
@flush_pager
@pass_controller
def version(controller):
    """Show version information."""
    include_all = controller.config['dev.catch_errors']
    click_echo(dob_versions(include_all=include_all))


# ***
# *** [LICENSE] Command.
# ***

@run.command(hidden=True, help=help_strings.LICENSE_HELP)
@show_help_finally
@flush_pager
def license():
    """Show license information."""
    _license()


def _license():
    """Show license information."""
    echo_license()


# ***
# *** [COPYRIGHT] Command.
# ***

@cmd_bunch_group_get_meta
@run.command(help=help_strings.COPYRIGHT_HELP)
@show_help_finally
@flush_pager
@pass_controller
def copyright(controller):
    """Display copyright information.."""
    echo_copyright()


@cmd_bunch_group_get_meta
@run.command(hidden=True, help=help_strings.ABOUT_COMMAND_HELP)
@show_help_finally
@flush_pager
@pass_controller
def about(controller):
    """Display copyright information.."""
    echo_copyright()


# ***
# *** [DETAILS] Command [about paths, config, etc.].
# ***

# MAYBE: (lb): Call this dob-show? dob-status? dob-info?
#   (2018-06-09: Trying aliases for now)
#   Some ideas: aliases=['show', 'status', 'info', 'config', 'details', 'appinfo'])
#   Though maybe 'show' should be alias for dob-current?
# MAYBE: Calling this 'appinfo' would make this command first in the --help....
#   @run.command(aliases=['show', 'status', 'info'], help=help_strings.DETAILS_HELP)
# See also similarly named commands that have generic meanings:
#   dob details | dob info | dob show
# i.e., details about what? info about what? showing what?
@cmd_bunch_group_get_meta
@run.command(aliases=['info'], help=help_strings.DETAILS_HELP)
@show_help_finally
@flush_pager
@click.option('--tmi', '--full', is_flag=True, help=help_strings.DETAILS_TMI_HELP)
@pass_controller
def details(controller, tmi):
    """List details about the runtime environment."""
    echo_app_details(controller, full=tmi)


# ***
# *** [ENVIRONS] Command [like details command, but shell-sourceable].
# ***

@cmd_bunch_group_get_meta
@run.command(aliases=['env'], help=help_strings.ENVIRONS_HELP)
@show_help_finally
@flush_pager
@pass_controller
def environs(controller):
    """List shell-sourceable details about the runtime environment."""
    echo_app_environs(controller)


# ***
# *** [DEBUG] Dump ya on a prompt.
# ***

@cmd_bunch_group_get_meta
@run.command(help=help_strings.DEBUG_HELP, hidden=True)
@show_help_finally
@flush_pager
@pass_controller
def debug(controller):
    """Break!"""
    import pdb
    pdb.set_trace()
    pass


# ***
# *** [DEMO] Command.
# ***

@cmd_bunch_group_introducing
@run.command('demo', help=help_strings.DEMO_HELP)
@show_help_finally
@flush_pager
@pass_controller
@demo_config
def demo_dob_and_nark(controller):
    """"""
    demo_dob(controller)


# ***
# *** [INIT] Command.
# ***

@cmd_bunch_group_introducing
@run.command('init', help=help_strings.INIT_HELP_OVERVIEW)
@show_help_finally
@flush_pager
@pass_controller
def init_config_and_store(controller):
    """"""
    controller.create_config_and_store(fact_cls=FactDressed)


# ***
# *** Shared Command Group settings.
# ***

run_group_kwargs = {
    # Groups should also use our class override so that usage is proper,
    # and other fancy features we add are part of group commands, too.
    'cls': ClickAliasableBunchyPluginGroup,
    # Ensure Click does not die on the -h/--help option, e.g.,
    # `dob config -h` should show help, not show "Error: Missing command."
    'invoke_without_command': True,
}


# ***
# *** [CONFIG] Commands.
# ***

@cmd_bunch_group_get_meta
@run.group('config', help=help_strings.CONFIG_GROUP_HELP, **run_group_kwargs)
@show_help_finally
@show_help_if_no_command
@flush_pager
@click.pass_context
def config_group(ctx):
    """Base `config` group command run prior to any of the dob-config commands."""
    pass  # The command group decorator prints help if no subcommand called.


# *** [CONFIG] CREATE

@config_group.command('create', aliases=['new'], help=help_strings.CONFIG_CREATE_HELP)
@show_help_finally
@flush_pager
@click.option('-f', '--force', is_flag=True,
              help=help_strings.CONFIG_CREATE_FORCE_HELP)
@pass_controller_context
@ensure_plugged_in
def config_create(ctx, controller, force):
    """"""
    controller.create_config(force)


# *** [CONFIG] SHOW

@config_group.command('show', aliases=['dump'], help=help_strings.CONFIG_SHOW_HELP)
@show_help_finally
@flush_pager
@cmd_options_output_format_any_input
@click.argument('section', nargs=1, default='')
@click.argument('keyname', nargs=1, default='')
@pass_controller_context
@ensure_plugged_in
def config_show(ctx, controller, section='', keyname='', **kwargs):
    """"""
    postprocess_options_output_format_any_input(kwargs)
    echo_config_table(controller, section, keyname, **kwargs)


# *** [CONFIG] EDIT

@config_group.command('edit', help=help_strings.CONFIG_EDIT_HELP)
@show_help_finally
@flush_pager
@pass_controller_context
@ensure_plugged_in
def config_edit(ctx, controller):
    """"""
    edit_config_file(controller)


# *** [CONFIG] GET

@config_group.command('get', aliases=['read'], help=help_strings.CONFIG_GET_HELP)
@show_help_finally
@flush_pager
@click.argument('parts', nargs=-1, metavar='[SECTION] KEYNAME VALUE')
@pass_controller_context
@ensure_plugged_in
def config_value_get(ctx, controller, parts):
    """"""
    echo_config_value(ctx, controller, parts)


# *** [CONFIG] SET

@config_group.command('set', aliases=['write'], help=help_strings.CONFIG_SET_HELP)
@show_help_finally
@flush_pager
@click.argument('parts', nargs=-1, metavar='[SECTION] KEYNAME VALUE')
@pass_controller_context
@ensure_plugged_in
def config_value_set(ctx, controller, parts):
    """"""
    write_config_value(ctx, controller, parts)


# *** [CONFIG] UPDATE

@config_group.command('update', help=help_strings.CONFIG_UPDATE_HELP)
@show_help_finally
@flush_pager
@pass_controller_context
@ensure_plugged_in
def config_update(ctx, controller):
    """Write "missing" key values to the user configuration file."""
    controller.round_out_config()


# ***
# *** [STYLES] Commands.
# ***

@cmd_bunch_group_personalize
@run.group('styles',
           help=help_strings.STYLES_GROUP_HELP,
           help_weight=10,
           **run_group_kwargs)
@show_help_finally
@show_help_if_no_command
@flush_pager
@click.pass_context
def styles_group(ctx):
    """Base `styles` group command run prior to any of the dob-styles commands."""
    pass  # The command group decorator prints help if no subcommand called.


# *** [STYLES] CREATE

@styles_group.command('create', aliases=['new'], help=help_strings.STYLES_CREATE_HELP)
@show_help_finally
@flush_pager
@cmd_options_styles_named
@click.option('-f', '--force', is_flag=True, help=help_strings.STYLES_CREATE_FORCE_HELP)
@pass_controller_context
@ensure_plugged_in
def styles_create(ctx, controller, name, force):
    """"""
    create_styles_conf(controller, name, force)


# *** [STYLES] CONF

@styles_group.command('conf', help=help_strings.STYLES_CONF_HELP)
@show_help_finally
@flush_pager
@cmd_options_styles_named
@cmd_options_styles_internal
# (lb): Cannot specify click.option name, and if just -a and --all, click uses
# all as the function parameter, which shadows the built-in. So add --complete.
@click.option('-a', '--complete', '--all', is_flag=True,
              help=_('Include all possible settings, even those not set by the style.'))
@pass_controller_context
@ensure_plugged_in
def styles_conf(ctx, controller, name, internal, complete):
    """"""
    echo_styles_conf(controller, name, internal, complete)


# *** [STYLES] EDIT

@styles_group.command('edit', help=help_strings.STYLES_EDIT_HELP)
@show_help_finally
@flush_pager
@pass_controller_context
@ensure_plugged_in
def styles_edit(ctx, controller):
    """"""
    edit_styles_conf(controller)


# *** [STYLES] LIST

@styles_group.command('list', help=help_strings.STYLES_LIST_HELP)
@show_help_finally
@flush_pager
@cmd_options_styles_internal
@pass_controller_context
@ensure_plugged_in
def styles_list(ctx, controller, internal):
    """"""
    echo_styles_list(controller, internal)


# *** [STYLES] SHOW

@styles_group.command('show', aliases=['dump'], help=help_strings.STYLES_SHOW_HELP)
@show_help_finally
@flush_pager
@cmd_options_styles_named
@cmd_options_output_format_any_input
@pass_controller_context
@ensure_plugged_in
def styles_show(ctx, controller, name, **kwargs):
    """"""
    postprocess_options_output_format_any_input(kwargs)
    echo_styles_table(controller, name, **kwargs)


# ***
# *** [RULES] Commands.
# ***

@cmd_bunch_group_personalize
@run.group('rules',
           help=help_strings.RULES_GROUP_HELP,
           help_weight=20,
           **run_group_kwargs)
@show_help_finally
@show_help_if_no_command
@flush_pager
@click.pass_context
def rules_group(ctx):
    """Base `rules` group command run prior to any of the dob-rules commands."""
    pass  # The command group decorator prints help if no subcommand called.


# *** [RULES] CREATE

@rules_group.command('create', aliases=['new'], help=help_strings.RULES_CREATE_HELP)
@show_help_finally
@flush_pager
@click.option('-f', '--force', is_flag=True, help=help_strings.RULES_CREATE_FORCE_HELP)
@pass_controller_context
@ensure_plugged_in
def rules_create(ctx, controller, force):
    """"""
    create_rules_conf(controller, force)


# *** [RULES] CONF

@rules_group.command('conf', help=help_strings.RULES_CONF_HELP)
@show_help_finally
@flush_pager
@cmd_options_rule_name
# (lb): Cannot specify click.option name, and if just -a and --all, click uses
# all as the function parameter, which shadows the built-in. So add --complete.
@click.option('-a', '--complete', '--all', is_flag=True,
              help=_('Include all possible settings, even those not set by the style.'))
@pass_controller_context
@ensure_plugged_in
def rules_conf(ctx, controller, name, complete):
    """"""
    echo_rules_conf(controller, name, complete)


# *** [RULES] EDIT

@rules_group.command('edit', help=help_strings.RULES_EDIT_HELP)
@show_help_finally
@flush_pager
@pass_controller_context
@ensure_plugged_in
def rules_edit(ctx, controller):
    """"""
    edit_rules_conf(controller)


# *** [RULES] LIST

@rules_group.command('list', help=help_strings.RULES_LIST_HELP)
@show_help_finally
@flush_pager
@pass_controller_context
@ensure_plugged_in
def rules_list(ctx, controller):
    """"""
    echo_rule_names(controller)


# *** [RULES] SHOW

@rules_group.command('show', aliases=['dump'], help=help_strings.RULES_SHOW_HELP)
@show_help_finally
@flush_pager
@cmd_options_rule_name
@cmd_options_output_format_any_input
@pass_controller_context
@ensure_plugged_in
def rules_show(ctx, controller, name, **kwargs):
    """"""
    postprocess_options_output_format_any_input(kwargs)
    echo_rules_table(controller, name, **kwargs)


# ***
# *** [IGNORE] Commands.
# ***

@cmd_bunch_group_personalize
@run.group('ignore',
           help=help_strings.IGNORE_GROUP_HELP,
           help_weight=30,
           **run_group_kwargs)
@show_help_finally
@show_help_if_no_command
@flush_pager
@click.pass_context
def ignore_group(ctx):
    """Base `ignore` group command run prior to any of the dob-ignore commands."""
    pass  # The command group decorator prints help if no subcommand called.


# *** [IGNORE] CREATE

@ignore_group.command('create', aliases=['new'], help=help_strings.IGNORE_CREATE_HELP)
@show_help_finally
@flush_pager
@click.option('-f', '--force', is_flag=True, help=help_strings.IGNORE_CREATE_FORCE_HELP)
@pass_controller_context
@ensure_plugged_in
def ignore_create(ctx, controller, force):
    """"""
    create_ignore_conf(controller, force)


# *** [IGNORE] EDIT

@ignore_group.command('edit', help=help_strings.IGNORE_EDIT_HELP)
@show_help_finally
@flush_pager
@pass_controller_context
@ensure_plugged_in
def ignore_edit(ctx, controller):
    """"""
    edit_ignore_file(controller)


# *** [IGNORE] LIST

@ignore_group.command('list', help=help_strings.IGNORE_LIST_HELP)
@show_help_finally
@flush_pager
@pass_controller_context
@ensure_plugged_in
def ignore_list(ctx, controller):
    """"""
    echo_ignore_sections(controller)


# *** [IGNORE] SHOW

@ignore_group.command('show', aliases=['dump'], help=help_strings.IGNORE_SHOW_HELP)
@show_help_finally
@flush_pager
@cmd_options_rule_name
@cmd_options_output_format_any_input
@pass_controller_context
@ensure_plugged_in
def ignore_show(ctx, controller, name, **kwargs):
    """"""
    postprocess_options_output_format_any_input(kwargs)
    echo_ignore_table(controller, name, **kwargs)


# ***
# *** [STORE] Commands.
# ***

# See also: 'migrate' commands. The store commands are mostly for initial setup.

@cmd_bunch_group_dbms
@run.group('store', help=help_strings.STORE_GROUP_HELP, **run_group_kwargs)
@show_help_finally
@show_help_if_no_command
@flush_pager
@click.pass_context
def store_group(ctx):
    """Base `store` group command run prior to dob-store commands."""
    pass


@store_group.command('create', aliases=['new'], help=help_strings.STORE_CREATE_HELP)
@show_help_finally
@flush_pager
@click.option('-f', '--force', is_flag=True,
              help=help_strings.STORE_CREATE_FORCE_HELP)
@pass_controller
def store_create(controller, force):
    """"""
    controller.create_data_store(force, fact_cls=FactDressed)


@store_group.command('path', help=help_strings.STORE_PATH_HELP)
@show_help_finally
@flush_pager
@pass_controller
def store_path(controller):
    """"""
    click_echo(controller.sqlite_db_path)


@store_group.command('url', help=help_strings.STORE_URL_HELP)
@show_help_finally
@flush_pager
@pass_controller
def store_url(controller):
    """"""
    click_echo(controller.data_store_url)


@store_group.command('upgrade-legacy', help=help_strings.STORE_UPGRADE_LEGACY_HELP)
@show_help_finally
@flush_pager
@click.argument('filename', nargs=1, type=click.File('r'), required=False)
@click.option('-f', '--force', is_flag=True,
              help=help_strings.STORE_UPGRADE_FORCE_HELP)
@pass_controller_context
@post_processor
def upgrade_legacy(ctx, controller, force, filename=None):
    """Migrate a legacy "Hamster" database."""
    # If filename is false, this method will raise.
    return upgrade_legacy_database_file(ctx, controller, file_in=filename, force=force)


# ***
# *** [STATS] Command.
# ***

@cmd_bunch_group_dbms
@run.command('stats', help=help_strings.STATS_HELP, help_weight=20)
@show_help_finally
@flush_pager
@pass_controller_context
@induct_newbies
def nark_stats(ctx, controller):
    """List stats about the user's data."""
    echo_data_stats(controller)


# ***
# *** [LIST] Commands.
# ***

@cmd_bunch_group_generate_report
# Use a command alias to avoid conflict with builtin of same name
# (i.e., we cannot declare this function, `def list()`).
@run.group(
    'list', help=help_strings.LIST_GROUP_HELP, help_weight=20, **run_group_kwargs,
)
@show_help_finally
@show_help_if_no_command
@flush_pager
@pass_controller_context
def list_group(ctx, controller):
    """Base `list` group command run prior to any of the dob-list commands."""
    # No-op; have already shown a help message.
    pass


# *** ACTIVITIES.

@list_group.command('activities', aliases=['activity', 'act'],
                    help=help_strings.LIST_ACTIVITIES_HELP)
@show_help_finally
@flush_pager
@cmd_options_any_search_query(command='list', item='activity', match=True, group=False)
@pass_controller_context
@induct_newbies
def list_activities(ctx, controller, *args, **kwargs):
    """List matching activities, filtered and sorted."""
    query_activities(ctx, controller, *args, **kwargs)


def query_activities(ctx, controller, *args, **kwargs):
    postprocess_options_normalize_search_args(kwargs)
    if kwargs['show_usage'] or kwargs['show_duration']:
        handler = usage_activity.usage_activities
    else:
        handler = list_activity.list_activities
    handler(controller, *args, **kwargs)


# *** CATEGORIES.

@list_group.command('categories', aliases=['category', 'cat'],
                    help=help_strings.LIST_CATEGORIES_HELP)
@show_help_finally
@flush_pager
@cmd_options_any_search_query(command='list', item='category', match=True, group=False)
@pass_controller_context
@induct_newbies
def list_categories(ctx, controller, *args, **kwargs):
    """List matching categories, filtered and sorted."""
    query_categories(ctx, controller, *args, **kwargs)


def query_categories(ctx, controller, *args, **kwargs):
    postprocess_options_normalize_search_args(kwargs)
    if kwargs['show_usage'] or kwargs['show_duration']:
        handler = usage_category.usage_categories
    else:
        handler = list_category.list_categories
    handler(controller, *args, **kwargs)


# *** TAGS.

@list_group.command('tags', aliases=['tag'], help=help_strings.LIST_TAGS_HELP)
@show_help_finally
@flush_pager
# TESTME/2020-05-16: group_target seems peculiar: it's not enabled on
# list-act or list-cat; but it could be interesting (to see what tags
# you use on which act@gories).
@cmd_options_any_search_query(command='list', item='tags', match=True, group=False)
@pass_controller_context
@induct_newbies
def list_tags(ctx, controller, *args, **kwargs):
    """List all tags, with filtering and sorting options."""
    query_tags(ctx, controller, *args, **kwargs)


def query_tags(ctx, controller, *args, **kwargs):
    postprocess_options_normalize_search_args(kwargs)
    if kwargs['show_usage'] or kwargs['show_duration']:
        handler = usage_tag.usage_tags
    else:
        handler = list_tag.list_tags
    handler(controller, *args, **kwargs)


# *** FACTS.

def _list_facts(controller, *args, cmd_journal=False, **kwargs):
    """Find matching facts, filtered and sorted."""
    postprocess_options_normalize_search_args(kwargs, cmd_journal=cmd_journal)
    list_fact.list_facts(controller, *args, **kwargs)


@list_group.command('facts', aliases=['fact'], help=help_strings.LIST_FACTS_HELP)
@show_help_finally
@flush_pager
# The `dob find` and `dob list fact` commands are the same.
@cmd_options_any_search_query(command='list', item='fact', match=True, group=True)
@pass_controller_context
@induct_newbies
def dob_list_facts(ctx, controller, *args, **kwargs):
    _list_facts(controller, *args, **kwargs)


# ***
# *** [SEARCH] Facts Command.
# ***

@cmd_bunch_group_generate_report
@run.command('find', aliases=['search'], help=help_strings.SEARCH_HELP, help_weight=5)
@show_help_finally
@flush_pager
# The `dob find` and `dob list fact` commands are the same.
@cmd_options_any_search_query(command='list', item='fact', match=True, group=True)
@pass_controller_context
@induct_newbies
def search_facts(ctx, controller, *args, **kwargs):
    _list_facts(controller, *args, **kwargs)


# ***
# *** [REPORT] Facts Command.
# ***

@cmd_bunch_group_generate_report
@run.command('report', help=help_strings.REPORT_HELP, help_weight=10)
@show_help_finally
@flush_pager
# The `dob report` command is `dob find` with specific defaults.
@cmd_options_any_search_query(command='journal', item='fact', match=True, group=True)
@pass_controller_context
@induct_newbies
def journal_report(ctx, controller, *args, **kwargs):
    # The journal command groups by Activity, Category, and Day by default,
    # unless the user specified a different grouping. Note that to apply no
    # grouping, the user would have to use the `dob find` command instead.
    _list_facts(controller, *args, cmd_journal=True, **kwargs)


# ***
# *** [USAGE] Commands.
# ***

@cmd_bunch_group_generate_report
@run.group(
    'usage', help=help_strings.USAGE_GROUP_HELP, help_weight=15, **run_group_kwargs,
)
@show_help_finally
@show_help_if_no_command
@flush_pager
@pass_controller_context
def usage_group(ctx, controller):
    """Base `usage` group command run prior to any of the dob-usage commands."""
    pass


# *** ACTIVITIES.

@usage_group.command('activities', aliases=['activity', 'act'],
                     help=help_strings.USAGE_ACTIVITIES_HELP)
@show_help_finally
@flush_pager
@cmd_options_any_search_query(command='usage', item='activity', match=True, group=False)
@pass_controller_context
@induct_newbies
def usage_activities(ctx, controller, *args, **kwargs):
    """List all activities. Provide optional filtering by name."""
    query_activities(ctx, controller, *args, **kwargs)


# *** CATEGORIES.

@usage_group.command('categories', aliases=['category', 'cat'],
                     help=help_strings.USAGE_CATEGORIES_HELP)
@show_help_finally
@flush_pager
@cmd_options_any_search_query(command='usage', item='category', match=True, group=False)
@pass_controller_context
@induct_newbies
def usage_categories(ctx, controller, *args, **kwargs):
    """List all categories. Provide optional filtering by name."""
    query_categories(ctx, controller, *args, **kwargs)


# *** TAGS.

@usage_group.command('tags', aliases=['tag'], help=help_strings.USAGE_TAGS_HELP)
@show_help_finally
@flush_pager
@cmd_options_any_search_query(command='usage', item='tags', match=True, group=False)
@pass_controller_context
@induct_newbies
def usage_tags(ctx, controller, *args, **kwargs):
    """List all tags' usage counts, with filtering and sorting options."""
    query_tags(ctx, controller, *args, **kwargs)


# *** FACTS.

@usage_group.command('facts', aliases=['fact'], help=help_strings.USAGE_FACTS_HELP)
@show_help_finally
@flush_pager
@cmd_options_any_search_query(command='usage', item='fact', match=True, group=True)
@pass_controller_context
@induct_newbies
def usage_facts(ctx, controller, *args, **kwargs):
    """List all tags' usage counts, with filtering and sorting options."""
    postprocess_options_normalize_search_args(kwargs)
    list_fact.list_facts(
        controller,
        *args,
        **kwargs,
    )


# ***
# *** [CURRENT-FACT] Commands: stop/cancel/current/latest/show.
# ***

@cmd_bunch_group_ongoing_fact
@run.command('cancel', help=help_strings.CANCEL_HELP)
@show_help_finally
@flush_pager
@click.option(
    '-f', '--force', '--purge', is_flag=True,
    help=_('Completely delete Fact, rather than just marking deleted.'),
)
@pass_controller_context
@induct_newbies
@post_processor
def cancel(ctx, controller, force):
    """Cancel the active Fact. Do not store the Fact in the backend."""
    return cancel_fact(controller, purge=force)


@cmd_bunch_group_ongoing_fact
@run.command('current', help=help_strings.CURRENT_HELP)
@show_help_finally
@flush_pager
@pass_controller_context
@induct_newbies
def current(ctx, controller):
    """Display the active Fact."""
    echo_ongoing_fact(controller)


@cmd_bunch_group_ongoing_fact
@run.command('latest', aliases=['last'], help=help_strings.LATEST_HELP)
@show_help_finally
@flush_pager
@pass_controller_context
@induct_newbies
def latest(ctx, controller):
    """Display the last saved Fact."""
    echo_latest_ended(controller)


@cmd_bunch_group_ongoing_fact
@run.command('show', help=help_strings.HELP_CMD_SHOW)
@show_help_finally
@flush_pager
@pass_controller_context
@induct_newbies
def show(ctx, controller):
    """Display the latest saved, or the active Fact."""
    echo_ongoing_or_ended(controller)


# ***
# *** [CREATE-FACT] Commands.
# ***

def generate_add_fact_command(time_hint):
    def _generate_add_fact_command(func):
        @cmd_options_fact_add
        @cmd_options_fact_dryable
        @pass_controller_context
        @induct_newbies
        @post_processor
        def _add_fact(ctx, controller, *args, editor, **kwargs):
            return add_fact(
                controller, *args, time_hint=time_hint, use_carousel=editor, **kwargs
            )
        return update_wrapper(_add_fact, func)
    return _generate_add_fact_command


@cmd_bunch_group_add_fact
@run.command("now", aliases=["now:"], help=help_strings.ADD_FACT_NOW)
@show_help_finally
@flush_pager
@cmd_options_factoid_verify_none
@generate_add_fact_command("verify_none")
def add_fact_now(controller, *args, **kwargs):
    """Start or add a fact using the `now` directive."""
    # Not reachable, because generate_add_fact_command.
    assert False  # pragma: no cover


@cmd_bunch_group_add_fact
@run.command("on", aliases=["on:"], help=help_strings.ADD_FACT_ON,
             hidden=True)  # FIXME/2019-11-21: Remove aliases, but keep unique help.
@show_help_finally
@flush_pager
@cmd_options_factoid_verify_none
@generate_add_fact_command("verify_none")
def add_fact_on(controller, *args, **kwargs):
    """Start or add a fact using the `on` directive."""
    # Not reachable, because generate_add_fact_command.
    assert False  # pragma: no cover


@cmd_bunch_group_add_fact
@run.command("at", aliases=["at:"], help=help_strings.ADD_FACT_AT)
@show_help_finally
@flush_pager
@cmd_options_factoid_verify_start
@generate_add_fact_command("verify_start")
def add_fact_at(controller, *args, **kwargs):
    """Start or add a fact using the `at` directive."""
    # Not reachable, because generate_add_fact_command.
    assert False  # pragma: no cover


@cmd_bunch_group_add_fact
@run.command("from", help=help_strings.ADD_FACT_FROM)
@show_help_finally
@flush_pager
@cmd_options_factoid_verify_both
@generate_add_fact_command("verify_both")
def add_fact_from(controller, *args, **kwargs):
    """Add a fact using the `from ... to/until` directive."""
    # Not reachable, because generate_add_fact_command.
    assert False  # pragma: no cover


@cmd_bunch_group_add_fact
@run.command("to", aliases=["to:"], help=help_strings.ADD_FACT_TO)
@show_help_finally
@flush_pager
@cmd_options_factoid_verify_end
@generate_add_fact_command("verify_end")
def add_fact_to(controller, *args, **kwargs):
    """Start or add a fact using the `to` directive."""
    # Not reachable, because generate_add_fact_command.
    assert False  # pragma: no cover


@cmd_bunch_group_add_fact
@run.command("until", aliases=["until:"], help=help_strings.ADD_FACT_UNTIL,
             hidden=True)  # FIXME/2019-11-21: Remove aliases... but I like "until"!
@show_help_finally
@flush_pager
@cmd_options_factoid_verify_end
@generate_add_fact_command("verify_end")
def add_fact_until(controller, *args, **kwargs):
    """Start or add a fact using the `until` directive."""
    # Not reachable, because generate_add_fact_command.
    assert False  # pragma: no cover


@cmd_bunch_group_add_fact
@run.command("then", aliases=["then:"], help=help_strings.ADD_FACT_THEN)
@show_help_finally
@flush_pager
# FIXME/2019-11-22: (lb): This right? (Because command is verify_then):
@cmd_options_factoid_verify_start
@generate_add_fact_command("verify_then")
def add_fact_then(controller, *args, **kwargs):
    """Start or add a fact using the `then` directive."""
    # Not reachable, because generate_add_fact_command.
    assert False  # pragma: no cover


@cmd_bunch_group_add_fact
@run.command("still", aliases=["still:"], help=help_strings.ADD_FACT_STILL)
@show_help_finally
@flush_pager
# FIXME/2019-11-22: (lb): This right? (Because command is verify_still):
@cmd_options_factoid_verify_start
@generate_add_fact_command("verify_still")
def add_fact_still(controller, *args, **kwargs):
    """Start or add a fact using the `still` directive."""
    # Not reachable, because generate_add_fact_command.
    assert False  # pragma: no cover


@cmd_bunch_group_add_fact
@run.command("after", aliases=["after:"], help=help_strings.ADD_FACT_AFTER)
@show_help_finally
@flush_pager
# FIXME/2019-11-22: (lb): This right? (Because command is verify_after):
@cmd_options_factoid_verify_start
@generate_add_fact_command("verify_after")
def add_fact_after(controller, *args, **kwargs):
    """Start or add a fact using the `after` directive."""
    # Not reachable, because generate_add_fact_command.
    assert False  # pragma: no cover


@cmd_bunch_group_add_fact
@run.command("next", aliases=["next:"], help=help_strings.ADD_FACT_NEXT,
             hidden=True)  # FIXME/2019-11-21: Remove aliases, but keep unique help.
@show_help_finally
@flush_pager
# FIXME/2019-11-22: (lb): This right? (Because command is verify_after):
@cmd_options_factoid_verify_start
# Note that 'next' is really just an alias of 'after'.
@generate_add_fact_command("verify_after")
def add_fact_next(controller, *args, **kwargs):
    """Start or add a fact using the `next` directive."""
    # Not reachable, because generate_add_fact_command.
    assert False  # pragma: no cover


# ***
# *** [START-STOP] Command(s).
# ***


# NOPE: @cmd_bunch_group_add_fact
@cmd_bunch_group_ongoing_fact
@run.command('start', help=help_strings.ADD_FACT_START, help_weight=5)
@show_help_finally
@flush_pager
@cmd_options_factoid_verify_none
@generate_add_fact_command('verify_none')
def add_fact_start(controller, *args, **kwargs):
    """Start or add a fact using the `now` directive."""
    # Not reachable, because generate_add_fact_command.
    assert False  # pragma: no cover


# NOPE: @cmd_bunch_group_add_fact
@cmd_bunch_group_ongoing_fact
@run.command('stop', help=help_strings.STOP_HELP, help_weight=5)
@show_help_finally
@flush_pager
@cmd_options_factoid_verify_end
# (lb) verify_end is imprecise, because stop should always set end time
# to now. At least if you're being pedantic. I just want to have a group
# of latest-fact commands that's hopefully simpler to grok than the more
# expansive group of add-fact commands.
@generate_add_fact_command('verify_end')
def stop(controller, *args, **kwargs):
    """Complete the active fact (by setting its 'end')."""
    # Not reachable, because generate_add_fact_command.
    assert False  # pragma: no cover


# ***
# *** [ADD-FACT-HELP] Command(s).
# ***


add_group_kwargs = {
    # Use a special business class that munges the usage to show all add commands.
    'cls': ClickAddFactHelpGroup,
    'invoke_without_command': True,
}


@cmd_bunch_group_add_fact
# - Hide this command; we'll show the misdirecting 'add --show' command instead,
#   but this command is necessary to catch `dob add --help` (i.e., catch the
#   command `add` with the option `--help`, as opposed to the next command,
#   which is the command `add --show` with no option, e.g., `dob "add --show"`.
# - ADD_FACT_GROUP_NAME is 'add', and a class member to automate usage building.
@run.group(
    ClickAddFactHelpGroup.ADD_FACT_GROUP_NAME,  # 'add'
    help=help_string_add_fact.ADD_FACT_COMMON,
    hidden=True,
    **add_group_kwargs
)
@show_help_finally
@show_help_if_no_command
@flush_pager
@cmd_options_factoid
@cmd_options_fact_add
@cmd_options_fact_dryable
@pass_controller_context
@induct_newbies
@post_processor
def add_group(ctx, controller, *args, **kwargs):
    """Base `add` group command run prior to any of the dob-add commands."""
    pass


@cmd_bunch_group_add_fact
# - (lb): Just a little "hack" (more like "misdirection") to make the first entry
#   in the add-fact bunchy group a hint to the user how to get help on adding Facts.
# - ADD_FACT_GROUP_HELP is 'add --help', and a class member to automate usage building.
@run.group(
    ClickAddFactHelpGroup.ADD_FACT_GROUP_HELP,  # 'add --help'
    help=help_string_add_fact.ADD_FACT_COMMON,
    **add_group_kwargs
)
@show_help_finally
@show_help_if_no_command
@flush_pager
@cmd_options_factoid
@cmd_options_fact_add
@cmd_options_fact_dryable
@pass_controller_context
@induct_newbies
@post_processor
def add_help(*args, **kwargs):
    """Misdirecting "add --help" command to improve general help."""
    pass


# ***
# *** [EDIT] Command(s).
# ***

@cmd_bunch_group_edit
@run.command('edit', help=help_strings.EDIT_FACT_HELP)
@show_help_finally
@flush_pager
@cmd_options_edit_item
@cmd_options_fact_edit
@pass_controller_context
@induct_newbies
@post_processor
def edit_group(ctx, controller, *args, **kwargs):
    """Base `edit` group command run prior to any of the dob-edit commands."""
    # NOTE: The return value is a list of edited Facts that's sent to the
    #       @post_processor decorator callbacks.
    return edit_fact_by_key(ctx, controller, *args, **kwargs)


def edit_fact_by_key(
    ctx,
    controller,
    *args,
    key,
    no_editor,
    edit_text,
    edit_meta,
    # Leave 'latest_1' in kwargs.
    **kwargs
):
    def _edit_fact_by_key():
        keys = assemble_keys()
        return process_edit_command(keys)

    def assemble_keys():
        keys = []
        if key is not None:
            keys.append(key)
        # See also if user specified '-1', '-2', etc.
        for kwg_key in kwargs.keys():
            match = re.match(r'^(latest_(\d+))$', kwg_key)
            if match is not None:
                # Check whether enabled (True) or not.
                if kwargs[match.groups()[0]]:
                    to_last_index = -1 * int(match.groups()[1])
                    keys.append(to_last_index)
        # (lb): If user runs plain `dob edit` be nice and assume latest Fact.
        #  - So `dob edit` same as `dob edit -1`.
        if not keys:
            keys = [-1]
        elif len(keys) > 1:
            dob_in_user_exit(_(
                "Argument error: Please specify “-1”, or Fact ID, not both!"
            ))
        return keys

    def process_edit_command(keys):
        edited_facts = edit_fact_by_pk(
            controller,
            key=keys[0],
            use_carousel=(not no_editor),
            edit_text=edit_text,
            edit_meta=edit_meta,
        )
        return edited_facts

    return _edit_fact_by_key()


# ***
# *** [EXPORT] Command.
# ***

@cmd_bunch_group_dbms
@run.command('export', help=help_strings.EXPORT_HELP)
@show_help_finally
@flush_pager
@cmd_options_any_search_query(command='export', item='fact', match=True, group=False)
@pass_controller_context
@induct_newbies
def transcode_export(ctx, controller, *args, **kwargs):
    """Export all facts of within a given time window to a file of specified format."""
    postprocess_options_normalize_search_args(kwargs)
    kwargs['output_format'] = 'factoid'
    list_fact.list_facts(controller, *args, **kwargs)


# ***
# *** [IMPORT] Command.
# ***

@cmd_bunch_group_dbms
@run.command('import', help=help_strings.IMPORT_HELP)
@show_help_finally
@flush_pager
@click.argument('filename', nargs=1, type=click.File('r'), required=False)
@click.option('-o', '--output', type=click.File('w', lazy=True),
              help=_('If specified, write to output file rather than saving'))
@click.option('-f', '--force', is_flag=True,
              help=_('Overwrite --output file if is exists'))
@click.option('-r', '--rule', '--sep', nargs=1, default='',
              help=_('With --output, split facts with a horizontal rule'))
@click.option('--backup/--no-backup', '-B', default=True, show_default=True,
              help=_('Keep plaintext backup of edited facts until committed'))
@click.option('-b', '--leave-backup', is_flag=True,
              help=_('Leave working backup file after commit'))
@cmd_options_fact_dryable
@cmd_options_fact_import
@pass_controller_context
@induct_newbies
@post_processor
def transcode_import(
    ctx,
    controller,
    output,
    force,
    no_editor,
    filename=None,
    *args,
    **kwargs
):
    """Import from file or STDIN (pipe).

    .. NOTE: You can get tricky and enter Facts on stdin using ``-``::

               dob import -

             Click will set filename to STDIN, and Dob process each line
             as the user types it (and presses ENTER), and stops on ^D.
    """

    # If neither file nor stdin specified, dump the help.
    if filename is None and sys.stdin.isatty():
        # There are two functionally equivalent ends to showing help.
        # - We could use the help helper with a string:
        #     help_command_help(ctx, ['import'])
        # - Or we can use the context to spill its help:
        click_echo(ctx.command.get_help(ctx))
        # We could `return` or `return []` here for post_processor, or exit 0
        # because we showed help and now we're done; but you can pipe stdin to
        # `dob import` and that is a real operation, so let's return 1, indicating
        # that maybe the command *is* wrong, because no input specified.
        sys.exit(1)

    # If opposite condition, complain -- user cannot send stdin and a filename.
    if filename is not None and not sys.stdin.isatty():
        # Use case:
        #   echo "2020-01-28 15:38: hello" | dob import facts.txt
        # Or:
        #   dob import - < facts.txt
        msg = 'Hey, why are you redirecting STDIN *and* specifying a file?'
        click_echo(msg)
        sys.exit(1)

    # If output file specified, verify file absent, or --force.
    if output and not force and os.path.exists(output.name):
        msg = _('Outfile already exists at: {}'.format(output.name))
        click_echo(msg)
        sys.exit(1)

    # If filename smells False, import_facts will use sys.stdin.
    # - FIXME/2019-11-19: Test not specifying file and see how it works on
    # its own vs. piped, e.g., test `dob import` vs. `cat file | dob import`.
    # NOTE: It use_carousel is True, user is forced to save through interface,
    #       and then quit, and then saved_facts is empty; but if the carousel
    #       was not used and the facts were immediately saved, saved_facts
    #       will be those facts, which are returned from this functions for
    #       @post_processor to use to call the plugins.
    pre_post_processed_saved_facts = import_facts(
        controller,
        *args,
        file_in=filename,
        file_out=output,
        use_carousel=(not no_editor),
        **kwargs
    )
    return pre_post_processed_saved_facts


# ***
# *** [COMPLETE] Command [Bash tab completion].
# ***

# FIXME: YAS! `hidden` is from a branch at:
#          sstaszkiewicz-copperleaf:6.x-maintenance
#        Watch the PR, lest you want to remove this before publishing:
#          https://github.com/pallets/click/pull/985
#          https://github.com/pallets/click/pull/500
@run.command('complete', hidden=True, help=help_strings.COMPLETE_HELP)
@show_help_finally
@flush_pager
@pass_controller_context
@induct_newbies
def complete(ctx, controller):
    """Bash tab-completion helper."""
    controller.disable_logging()
    tab_complete(controller)


# ***
# *** [MIGRATE] Commands [database transformations].
# ***

@cmd_bunch_group_dbms
@run.group(
    'migrate',
    help=help_strings.MIGRATE_GROUP_HELP,
    # (lb): Hidden until needed. Does db really need to evolve?
    # Then again, I don't want to be rigid like JSON or Md specs.
    # I'm open to fresh ideas.
    hidden=True,
    **run_group_kwargs,
)
@show_help_finally
@show_help_if_no_command
@flush_pager
@pass_controller_context
def migrate_group(ctx, controller):
    """Base `migrate` group command run prior to any of the dob-migrate commands."""
    pass


@migrate_group.command('control', help=help_strings.MIGRATE_CONTROL_HELP)
@show_help_finally
@flush_pager
@pass_controller_context
@insist_germinated
def _migrate_control(ctx, controller):
    """Mark a database as under version control."""
    migrate_control(controller)


@migrate_group.command('down', help=help_strings.MIGRATE_DOWN_HELP)
@show_help_finally
@flush_pager
@pass_controller_context
@insist_germinated
def _migrate_downgrade(ctx, controller):
    """Downgrade the database according to its migration version."""
    migrate_downgrade(controller)


@migrate_group.command('up', help=help_strings.MIGRATE_UP_HELP)
@show_help_finally
@flush_pager
@pass_controller_context
@insist_germinated
def _migrate_upgrade(ctx, controller):
    """Upgrade the database according to its migration version."""
    migrate_upgrade(controller)


@migrate_group.command('version', help=help_strings.MIGRATE_VERSION_HELP)
@show_help_finally
@flush_pager
@pass_controller_context
@insist_germinated
def _migrate_version(ctx, controller):
    """Show migration information about the database."""
    migrate_version(controller)


# 2018-07-15 14:00ish: To loaded: 0.440 secs.
# After adding lazy-loading:
# 2018-07-15 18:06:    To loaded: 0.060 secs.
# 2018-07-15 18:06:    To run:    0.329 secs.
# About ~ 0.150 is sqlalchemy, which is unavoidable?
# About ~ 0.005 is loading config.
# The other half of the time, ~ 0.145, is plugins.
profile_elapsed('To dob: loaded')




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

import sys


__all__ = (
    'prompt_and_save_confirmed',
)


# ***

def prompt_and_save_confirmed(
    controller,
    edit_facts=None,
    orig_facts=None,
    use_carousel=False,
    backup_callback=None,
    file_out=None,
    rule='',
    yes=False,
    dry=False,
    progress=None,
    **kwargs,
):
    """"""

    def _prompt_and_save_confirmed():
        saved_facts = []
        if not edit_facts:
            return saved_facts
        must_isatty_unless_testing()
        saved_facts = launch_carousel_or_prompt_directly()
        return saved_facts

    def must_isatty_unless_testing():
        if (
            not use_carousel
            or sys.stdin.isatty()
            or 'input' in kwargs  # PTK3's input test hook.
        ):
            return
        raise Exception(_(
            'Commands requires user confirmation, or --yes or --dry.'
        ))

    def launch_carousel_or_prompt_directly():
        if use_carousel:
            return launch_carousel()
        else:
            return prompt_directly()

    def launch_carousel():
        # Not just lazy loading, but allows test_save_backedup to mock away.
        from dob_viewer.traverser.save_confirmer import prompt_and_save_confirmer
        prompt_and_save_confirmer(
            controller,
            edit_facts=edit_facts,
            orig_facts=orig_facts,
            backup_callback=backup_callback,
            dry=dry,
            **kwargs,
        )
        return []

    def prompt_directly():
        # Not just lazy loading, but allows test_save_backedup to mock away.
        from .save_confirmer import prompt_and_save_confirmer
        saved_facts = prompt_and_save_confirmer(
            controller,
            edit_facts=edit_facts,
            orig_facts=orig_facts,
            backup_callback=backup_callback,
            file_out=file_out,
            rule=rule,
            yes=yes,
            dry=dry,
            progress=progress,
            **kwargs,
        )
        return saved_facts

    # ***

    return _prompt_and_save_confirmed()




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma, © 2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

import click_hotoffthehamster as click

from dob_bright.reports.tabulate_results import report_table_columns
from dob_bright.termio import dob_in_user_exit, dob_in_user_warning

__all__ = (
    # One decorator is all you need for each list and usage command.
    'cmd_options_any_search_query',
    # Some other commands share use of the ASCII output table feature.
    'cmd_options_output_format_any_input',
    'cmd_options_output_format_facts_only',
    # Argument parsing helpers, to facilitate **kwargs passing.
    'postprocess_options_normalize_search_args',
    'postprocess_options_output_format_any_input',
    # Private module variables:
    #   '_cmd_options_*',
    # Private module functions:
    #   '_cmd_options_output_format_choices',
    #   '_cmd_options_output_format_multiple_choices_option',
    #   '_cmd_options_output_format_singular_options_any',
    #   '_cmd_options_output_format_singular_options_fact',
    #   '_cmd_options_output_format_tabling',
    #   '_cmd_options_output_formats_basic',
    #   '_cmd_options_output_formats_table',
    #   '_cmd_options_results_group',
    #   '_cmd_options_results_output_path',
    #   '_cmd_options_results_sort_order',
    #   '_cmd_options_search_time_window',
    #   '_postprocess_options_grouping',
    #   '_postprocess_options_match_activities',
    #   '_postprocess_options_match_categories',
    #   '_postprocess_options_match_tags',
    #   '_postprocess_options_matching',
    #   '_postprocess_options_sort_dirs',
    #   '_postprocess_options_sort_cols',
    #   '_postprocess_options_results_show_hide',
    #   '_postprocess_options_results_show_hide_option',
    #   '_postprocess_options_search_term',
    #   '_postprocess_options_sparkline',
    #   '_postprocess_options_sparkline_float',
    #   '_postprocess_options_sparkline_secs',
    #   '_postprocess_options_sparkline_total',
)


# ***
# *** [SEARCH QUERY] Item ID.
# ***

_cmd_options_search_item_key = [
    click.option(
        '-k', '--key',
        metavar='ID',
        help=_('The database key of the item.'),
    ),
]


# ***
# *** [SEARCH QUERY] Search Terms.
# ***

_cmd_options_search_search_term = [
    click.argument(
        'search_term',
        nargs=-1,
        default=None,
    ),
]


def _postprocess_options_search_term(kwargs):
    if 'search_term' not in kwargs:
        return

    # (lb): Me, the pedant.
    kwargs['search_terms'] = kwargs['search_term']
    del kwargs['search_term']


# ***
# *** [SEARCH QUERY] Time Window.
# ***

def _cmd_options_search_time_window(command=''):
    since_default_value = ''
    since_show_default = False
    if command == 'journal':
        since_default_value = 'last week'
        since_show_default = True
    cmd_options = [
        click.option(
            '-s', '--since', '--after',
            metavar='TIME',
            default=since_default_value,
            show_default=since_show_default,
            help=_('Show items newer than a specific date.'),
        ),
        click.option(
            '-u', '--until', '--before',
            metavar='TIME',
            help=_('Show items older than a specific date.'),
        ),
    ]
    return cmd_options


# ***
# *** [SEARCH QUERY] Deleted and Hidden.
# ***

# FIXME/2020-05-16: (lb): Cleanup these options upon broader cleanup.

_cmd_options_search_deleted_hidden = [
    click.option(
        '--deleted', is_flag=True, help=_('Show deleted items.'),
    ),
    # FIXME/2020-05-20: Scrub hidden from everywhere.
    # click.option(
    #     '--hidden', is_flag=True, help=_('Show hidden items.'),
    # ),
]


# ***
# *** [SEARCH MATCH] Activity.
# ***

_cmd_options_search_match_activities = [
    click.option(
        '-a', '--activity', multiple=True,
        help=_('Restrict results by exact activity name(s).'),
    ),
]


def _postprocess_options_match_activities(kwargs):
    if 'activity' not in kwargs:
        return

    match_activities = _postprocess_assemble_match_names(kwargs['activity'])
    if match_activities:
        kwargs['match_activities'] = match_activities
    del kwargs['activity']


def _postprocess_assemble_match_names(match_names):
    match_list = []
    for name in match_names:
        match_list.append(name)
        if not name:
            # Match empty string or NULL (this is a special case,
            # otherwise user has no way to match NULL from CLI).
            match_list.append(None)
    return match_list


# ***
# *** [SEARCH MATCH] Category.
# ***

_cmd_options_search_match_categories = [
    click.option(
        '-c', '--category', multiple=True,
        help=_('Restrict results by exact category name(s).'),
    ),
]


def _postprocess_options_match_categories(kwargs):
    if 'category' not in kwargs:
        return

    match_categories = _postprocess_assemble_match_names(kwargs['category'])
    if match_categories:
        kwargs['match_categories'] = match_categories
    del kwargs['category']


# ***
# *** [SEARCH MATCH] Tag names.
# ***

_cmd_options_search_match_tags = [
    click.option(
        '-t', '--tag', multiple=True,
        help=_('Restrict results by exact tag name(s).'),
    ),
]


def _postprocess_options_match_tags(kwargs):
    if 'tag' not in kwargs:
        return

    match_tags = _postprocess_assemble_match_names(kwargs['tag'])
    if match_tags:
        kwargs['match_tags'] = match_tags
    del kwargs['tag']


# ***
# *** [SEARCH MATCH] Broad matching.
# ***

_cmd_options_search_broad_match = [
    click.option(
        '--broad-match', '--broad',
        is_flag=True,
        help=_(
            'Try SEARCH_TERM matching on activity, category, and tag names'
            ' (otherwise SEARCH_TERM only matches description).'
        ),
    ),
]


# ***
# *** [RESULTS GROUP] Option.
# ***

def _cmd_options_results_group(item):
    choices = []
    if item != 'activity':
        choices.append('activity')
    if item != 'category':
        choices.append('category')
    if item != 'tags':
        choices.append('tags')
    if item == 'fact':
        choices.append('days')

    return [
        click.option(
            '-g', '--group', multiple=True,
            type=click.Choice(choices),
            help=_('Alias of `--group-<attribute>` options (multiple allowed).'),
        ),
    ]


def _postprocess_options_grouping(kwargs, cmd_journal=False):
    def __postprocess_options_grouping():
        if 'group' not in kwargs:
            return

        process_grouping_option('activity')
        process_grouping_option('category')
        process_grouping_option('tags')
        process_grouping_option('days')
        del kwargs['group']

        ensure_default_grouping()

    def process_grouping_option(type_name):
        # Form the group name, e.g., group_activity, group_category, etc.
        group_name = 'group_{}'.format(type_name)
        if group_name not in kwargs:
            return

        kwargs[group_name] = kwargs[group_name] or type_name in kwargs['group']

    def ensure_default_grouping():
        if not cmd_journal:
            return

        if not (
            kwargs['group_activity']
            or kwargs['group_category']
            or kwargs['group_tags']
            or kwargs['group_days']
        ):
            kwargs['group_activity'] = True
            kwargs['group_category'] = True
            kwargs['group_days'] = True

    __postprocess_options_grouping()


# ***
# *** [RESULTS GROUP] Activity.
# ***

_cmd_options_results_group_activity = [
    click.option(
        '-A', '--group-activity', is_flag=True,
        help=_('Group results by activity name.'),
    ),
]


# ***
# *** [RESULTS GROUP] Category.
# ***

_cmd_options_results_group_category = [
    # There's a global -C/--config option, but globals need to be
    # specified before the command name (just how Click works), so
    # this re-usage of the same single -C option is perfectly fine.
    click.option(
        '-C', '--group-category', is_flag=True,
        help=_('Group results by category name.'),
    ),
]


# ***
# *** [RESULTS GROUP] Tags.
# ***

_cmd_options_results_group_tags = [
    click.option(
        '-T', '--group-tags', is_flag=True,
        help=_('Group results by tag names.'),
    ),
]


# ***
# *** [RESULTS GROUP] Days.
# ***

_cmd_options_results_group_days = [
    click.option(
        '-Y', '--group-days', is_flag=True,
        help=_('Group results by day.'),
    ),
]


# ***
# *** [SEARCH RESULTS] Order.
# ***

def _cmd_options_results_sort_order(item, command, group):
    choices = ['name', 'activity', 'category']
    default_sort_cols = ['name']
    default_sort_orders = ['asc']
    if item == 'fact' or command == 'usage':
        choices += ['start', 'time']
        if item == 'fact':
            default_sort_cols = ['start']
        if group or command == 'usage':
            choices += ['usage']
            default_sort_cols = ['usage']
            default_sort_orders = ['desc']
        if command == 'journal':
            default_sort_cols = ['day', 'time']
            default_sort_orders = ['asc', 'desc']

    if group and item == 'fact':
        choices.append('day')
    if item in ('tag', 'fact'):
        # For item == 'tag', this option same as --sort 'name'.
        choices.append('tag')
    if item == 'fact':
        # Sorts by Fact PK. (lb): Not sure how useful.
        # - Because Momentaneous Facts, potentially useful.
        choices.append('fact')

    return [
        click.option(
            '-S', '--sort', '--order',
            default=default_sort_cols,
            show_default=True,
            type=click.Choice(choices),
            multiple=True,
            help=_('Order by column(s) (multiple allowed).'),
        ),
        click.option(
            '-D', '--direction', '--dir',
            default=default_sort_orders,
            type=click.Choice(['asc', 'desc']),
            multiple=True,
            help=_('Order by direction(s) (one for each --sort).'),
        ),
    ]


# ***
# *** [POST PROCESS] Sort/Order Options.
# ***


def _postprocess_options_sort_cols(kwargs):
    if 'sort' not in kwargs:
        return

    if kwargs['sort']:
        kwargs['sort_cols'] = kwargs['sort']
    del kwargs['sort']


def _postprocess_options_sort_dirs(kwargs):
    if 'direction' not in kwargs:
        return

    last_direction = 'asc'
    sort_orders = []
    if 'sort_cols' in kwargs:
        # Ensure sort_orders same length as sort_cols.
        for idx in range(len(kwargs['sort_cols'])):
            try:
                last_direction = kwargs['direction'][idx]
            except IndexError:
                pass
            sort_orders.append(last_direction)
    del kwargs['direction']

    if sort_orders:
        kwargs['sort_orders'] = sort_orders


# ***
# *** [DOUBLE-SORT] Developer Option.
# ***

_postprocess_options_results_options_sort_double_sort = [
    click.option(
        '--re-sort',
        is_flag=True,
        hidden=True,
        help=_('Resort query after SELECT post-processing.'),
    ),
]


# ***
# *** [SEARCH RESULTS] Limit and Offset.
# ***

_cmd_options_search_limit_offset = [
    click.option(
        '-L', '--limit', default=0, show_default=False,
        help=_('Limit the number of records to fetch.'),
    ),
    click.option(
        '-O', '--offset', default=0, show_default=False,
        help=_('Skip this number of records before returning results.'),
    ),
]


# ***
# *** [RESULTS HIDE] Description.
# ***

_cmd_options_results_hide_description = [
    click.option(
        '-P', '--hide-description', is_flag=True,
        help=_('Omit Fact description from results.'),
    ),
]


# ***
# *** [RESULTS SHOW/HIDE] Duration.
# ***

_cmd_options_results_hide_duration = [
    click.option(
        '-N', '--hide-duration', is_flag=True,
        help=_('Omit duration from results.'),
    ),
]


_cmd_options_results_show_duration = [
    click.option(
        '-N', '--show-duration', is_flag=True,
        help=_('Display duration time in results.'),
    ),
]


# ***
# *** [RESULTS HIDE/SHOW] Usage.
# ***

_cmd_options_results_hide_usage = [
    click.option(
        '-U', '--hide-usage', is_flag=True,
        help=_('Omit usage count from results.'),
    ),
]


_cmd_options_results_show_usage = [
    click.option(
        '-U', '--show-usage', is_flag=True,
        help=_('Display usage count in results.'),
    ),
]


# ***
# *** [POST PROCESS] Show/Hide.
# ***

def _postprocess_options_results_show_hide(kwargs):
    # The list command have --show-* options; the usage commands
    # have --hide-* options; the export command has neither.
    _postprocess_options_results_show_hide_option(kwargs, 'duration')
    _postprocess_options_results_show_hide_option(kwargs, 'usage')


def _postprocess_options_results_show_hide_option(kwargs, argname):
    show_item = None
    attr_show = 'show_{}'.format(argname)
    attr_hide = 'hide_{}'.format(argname)
    if attr_show in kwargs:
        show_item = kwargs[attr_show]
    elif attr_hide in kwargs:
        show_item = not kwargs[attr_hide]
    if show_item is not None:
        kwargs[attr_show] = show_item
    if attr_hide in kwargs:
        del kwargs[attr_hide]


# ***
# *** [RESULTS CUSTOM] Columns.
# ***

_cmd_options_results_show_columns = [
    click.option(
        '-l', '--column', multiple=True,
        type=click.Choice(report_table_columns()),
        help=_('Specify custom report columns.'),
    ),
]


# ***
# *** [SEARCH RESULTS] Table/Line width constraint.
# ***

_cmd_options_results_max_width = [
    click.option(
        '-W', '--max-width', '--width', default=-1,
        help=_('Set the table width or Factoid line length.'),
    ),
]


# ***
# *** [SEARCH RESULTS] Output file.
# ***

def _cmd_options_results_output_path(for_export=False):
    return [
        click.option(
            '-o', '--output',
            # Not setting `type=click.File('r')` because
            # refuses empty string, which we want user to
            # be able to use to choose stdout (i.e., when
            # for_export=True).
            help=_('Write to file instead of stdout.'),
            metavar='PATH',
            required=for_export,
        ),
    ]


def _postprocess_options_output_filename(kwargs):
    if 'output' not in kwargs:
        return

    kwargs['output_path'] = kwargs['output']
    del kwargs['output']


# ***
# *** [REPORT FORMAT] Basic Formats.
# ***


_standard_formats = [
    'csv',
    'json',
    'tsv',
    'xml',
    'table',
]


_single_fact_formats = [
    'factoid',
    'ical',
]


_fact_usage_formats = [
    'journal',
]


def _cmd_options_output_formats_basic(item=''):
    format_choices = []
    format_choices += _fact_usage_formats if item == 'fact' else []
    format_choices += _single_fact_formats if item == 'fact' else []
    format_choices += _standard_formats
    return format_choices


def cmd_options_output_format_facts_only():
    return _single_fact_formats


def _cmd_options_output_format_singular_options_any():
    # --format <format> aliases.
    cmd_options = [
        click.option(
            '--csv', is_flag=True,
            help=_('Output results as comma-separated values (CSV).'),
        ),
        click.option(
            '--json', '-J', is_flag=True,
            help=_('Output results as JavaScript Object Notation (JSON).'),
        ),
        click.option(
            '--tsv', is_flag=True,
            help=_('Output results as tab-separated values (TSV).'),
        ),
        click.option(
            '--xml', is_flag=True,
            help=_('Output results as Extensible Markup Language (XML).'),
        ),
        click.option(
            '--table', is_flag=True,
            help=_('Output results as format indicated by --table-type.'),
        ),
    ]
    return cmd_options


def _cmd_options_output_format_singular_options_fact():
    cmd_options = [
        click.option(
            '--journal', is_flag=True,
            help=_("Output Facts using dob's Journal format."),
        ),
        click.option(
            '--factoid', is_flag=True,
            help=_("Output Facts using dob's Factoid format."),
        ),
        click.option(
            '--ical', is_flag=True,
            help=_('Output Facts using iCalendar format.'),
        ),
    ]
    return cmd_options


def _cmd_options_output_format_multiple_choices_option(command='', item=''):
    format_choices = _cmd_options_output_formats_basic(item)

    if command == 'journal':
        default_format = 'journal'
    else:
        default_format = 'table'

    cmd_options = [
        click.option(
            '-f', '--format',
            type=click.Choice(format_choices),
            default=default_format,
            show_default=True,
            help=_('Alias of format options (one allowed).'),
        ),
    ]
    return cmd_options


def _cmd_options_output_format_choices(command='', item=''):
    cmd_options = []
    if item == 'fact':
        cmd_options += _cmd_options_output_format_singular_options_fact()
    cmd_options += _cmd_options_output_format_singular_options_any()
    cmd_options += _cmd_options_output_format_multiple_choices_option(command, item)
    return cmd_options


def _postprocess_options_output_format_choices(kwargs):
    if 'format' not in kwargs:
        return

    format_choices = _cmd_options_output_formats_basic(item='fact')
    fmts_specified = []
    for choice in format_choices:
        try:
            if kwargs[choice]:
                kwargs['format'] = choice
                fmts_specified.append(choice)
            del kwargs[choice]
        except KeyError:
            # 'choice' is 'factoid' or 'journal', because item=='fact' unknown, so
            # always trying, even for commands that do not output Fact reports.
            pass

    if len(fmts_specified) > 1:
        dob_in_user_warning(_(
            'More than one format specified: {}'
        ).format(fmts_specified))

    kwargs['output_format'] = kwargs['format']
    del kwargs['format']


# ***
# *** [REPORT FORMAT] Table Formats.
# ***

_tabulate_tablefmts_markup = [
    'github',
    #   | Header 1   | Header 2   | Header 3   |
    #   |------------|------------|------------|
    #   | value 1    | value 2    | value 3    |

    'html',
    #   <table>
    #   <thead>
    #   <tr><th>Header 1  </th><th>Header 2  </th><th>Header 3  </th></tr>
    #   </thead>
    #   <tbody>
    #   <tr><td>value 1   </td><td>value 2   </td><td>value 3   </td></tr>
    #   </tbody>
    #   </table>

    # 'jira',  # MEH: Could add if requested.
    #   || Header 1   || Header 2   || Header 3   ||
    #   | value 1    | value 2    | value 3    |

    # 'latex',  # MEH: Could add if requested.
    #   \begin{tabular}{lll}
    #   \hline
    #    Header 1   & Header 2   & Header 3   \\
    #   \hline
    #    value 1    & value 2    & value 3    \\
    #   \hline
    #   \end{tabular}

    # 'latex_raw',  # SKIP: Same as 'latex'
    #   \begin{tabular}{lll}
    #   \hline
    #    Header 1   & Header 2   & Header 3   \\
    #   \hline
    #    value 1    & value 2    & value 3    \\
    #   \hline
    #   \end{tabular}

    # 'latex_booktabs',  # MEH: Could add if requested.
    #   \begin{tabular}{lll}
    #   \toprule
    #    Header 1   & Header 2   & Header 3   \\
    #   \midrule
    #    value 1    & value 2    & value 3    \\
    #   \bottomrule
    #   \end{tabular}

    'mediawiki',
    #   {| class="wikitable" style="text-align: left;"
    #   |+ <!-- caption -->
    #   |-
    #   ! Header 1   !! Header 2   !! Header 3
    #   |-
    #   | value 1    || value 2    || value 3
    #   |}

    # 'moinmoin',  # For MoinMoin Wiki Engine. MEH: Could add if requested.
    #   || ''' Header 1   ''' || ''' Header 2   ''' || ''' Header 3   ''' ||
    #   ||  value 1     ||  value 2     ||  value 3     ||

    'orgtbl',
    #   | Header 1   | Header 2   | Header 3   |
    #   |------------+------------+------------|
    #   | value 1    | value 2    | value 3    |

    'rst',
    #   ==========  ==========  ==========
    #   Header 1    Header 2    Header 3
    #   ==========  ==========  ==========
    #   value 1     value 2     value 3
    #   ==========  ==========  ==========

    # 'textile',  # MEH: Could add if requested.
    #   |_.  Header 1   |_. Header 2   |_. Header 3   |
    #   |<. value 1     |<. value 2    |<. value 3    |

    # 'unsafehtml',  # MEH.
    #   Header 1    Header 2    Header 3
    #   ----------  ----------  ----------
    #   value 1     value 2     value 3

    # 'youtrack',  # Similar to 'jira'
    #   ||  Header 1    ||  Header 2    ||  Header 3    ||
    #   |  value 1     |  value 2     |  value 3     |
]


# (lb): I tested 3 ASCII table packages: texttable, tabulate, and humanfriendly.
# - The texttable package (as far as I am aware) is the only library that wraps
#   cell values automatically. So this is the package we use to generate an ASCII
#   table that is meant to be displayed in the user's terminal.
# - We omit the humanfriendly table generator because it does not wrap column
#   values automatically, and does not offer anything that texttable does not,
#   other than perhaps slightly differently table borders.
# - The tabulate package has a number of seemingly interesting table formats that
#   the user could choose (like 'fancy_grid', 'grid', 'pretty', 'simple', etc.),
#   but it also does not wrap column values automatically, so really none of these
#   formats is useable (and none are essential, either; they're just stylistic).
# - As such, if we wanted to use tabulate or friendly to generate a readable (in the
#   user's terminal) ASCII table, we'd have to do the column value wrapping ourself.
#   For instance, tabulate will wrap column values, but only on existing newlines, so
#   we'd have to, say, determine the appropriate column widths first, and then call
#   ansiwrap.wrap on all the values before calling tabulate.
#   - Without the upfront wrapping, the tabulate and humanfriendly packages print
#     the table as wide as the column values dictate, which can cause the table
#     rows to bleed across lines, which makes the table impractical to read.
# - So, at least for generating a useable ASCII table, we use texttable, and we
#   make texttable the default table choice, although we label it the 'normal'
#   table type, so that the user is not distracted by package naming minutia.
# - However, we can expose a few tabulate format options that are not meant
#   to be displayed or read in the terminal, such as HTML and reST formats.
# - tl;dr We let the user choose `texttable` to generate an ASCII table; and we
#   let them choose a few of the `tabulate` formats to generate alternative
#   (non-table) outputs.
def _cmd_options_output_formats_table(item=''):
    table_choices = []
    # MAGIC_VALUE: Use 'normal' to refer to the nice, wrapped ASCII table
    #              that 'texttable' generates by default.
    table_choices += ['normal']
    # Include also those tabulate package output formats that are not
    # ASCII table formats (and not destined for terminal viewage).
    table_choices += _tabulate_tablefmts_markup
    return table_choices


def _cmd_options_output_format_tabling():
    formats_table = _cmd_options_output_formats_table()

    cmd_options = [
        click.option(
            '--table-type', '--type',
            type=click.Choice(formats_table),
            nargs=1,
            default='normal',
            show_default=True,
            help=_('Table format style.'),
        ),
    ]
    return cmd_options


def _postprocess_options_output_format_tabling(kwargs):
    if 'table_type' not in kwargs:
        return

    if kwargs['table_type'] == 'normal':
        kwargs['table_type'] = 'texttable'


# ***
# *** [REPORT FORMAT] Basic Formats and Table Styles.
# ***

def cmd_options_output_format_any_input(func):
    cmd_options = []
    cmd_options.extend(_cmd_options_output_format_choices())
    cmd_options.extend(_cmd_options_output_format_tabling())
    for option in reversed(cmd_options):
        func = option(func)
    return func


def postprocess_options_output_format_any_input(kwargs, cmd_journal=False):
    _postprocess_options_output_format_choices(kwargs)
    _postprocess_options_output_format_tabling(kwargs)


# ***
# *** [EXPORT FORMAT] Factoid Option.
# ***

_cmd_options_output_factoids_hrule = [
    click.option(
        # MEH: We could alias '-R', or reserve -R for later.
        '--factoid-rule', '--rule',
        nargs=1,
        default='',
        metavar='SEP',
        help=_('Separate Factoids with a horizontal rule.'),
    ),
]


# ***
# *** [REPORT/EXPORT/OUTPUT FORMAT] Sparkline Options.
# ***

_cmd_options_output_sparkline_format = [
    click.option(
        # MEH: We could alias '-W', or reserve -W for later.
        '--spark-width', '--swid',
        type=int,
        default='12',
        metavar='INT',
        # (lb): I like the idea of showing "block (█) character"
        #       but it's very distracting.
        help=_("Number of block characters in full spark [default: 12]."),
    ),
    click.option(
        # MEH: We could alias '-V', or reserve -V for later.
        '--spark-total', '--stot',
        nargs=1,
        default='max',
        metavar='MAX',
        # show_default=True,
        help=_(
            "Number of seconds in full spark: "
            "'max' or 'net' duration, or number [default: 'max']."
        ),
    ),
    click.option(
        # MEH: We could alias '-S', or reserve -S for later.
        '--spark-secs', '--ssec',
        metavar='WIDTH',
        help=_("Seconds per block character [default: total / width]."),
    ),
]


# ***
# *** [POST PROCESS] Show/Hide.
# ***

def _postprocess_options_sparkline(kwargs):
    if 'spark_total' not in kwargs:
        return

    _postprocess_options_sparkline_total(kwargs)
    _postprocess_options_sparkline_secs(kwargs)


def _postprocess_options_sparkline_total(kwargs):
    if kwargs['spark_total'] in ('max', 'net'):
        return

    _postprocess_options_sparkline_float(kwargs, 'spark_total')


def _postprocess_options_sparkline_secs(kwargs):
    _postprocess_options_sparkline_float(kwargs, 'spark_secs')


def _postprocess_options_sparkline_float(kwargs, spark_attr):
    try:
        kwargs[spark_attr] = float(kwargs[spark_attr] or 0)
    except ValueError:
        try:
            # Is this a security issue? Ha!
            # - Let user specify math on the command line, e.g., to
            #   specify 8 hours as the full spark width, user'd use:
            #     --spark-total '8 * 60 * 60'
            kwargs[spark_attr] = float(eval(kwargs[spark_attr]))
        except Exception:
            msg = _(
                "Unable to parse --{} value as (eval'able) seconds: {}"
            ).format(spark_attr.replace('_', '-'), kwargs[spark_attr])
            dob_in_user_exit(msg)


# ***
# *** [RESULTS OUTPUT] Aggregate Totals.
# ***

_cmd_options_output_show_totals = [
    click.option(
        '--show-totals', '--totals',
        is_flag=True,
        help=_('Calculate and show totals in final row of output.'),
    ),
]


_cmd_options_output_hide_totals = [
    click.option(
        '--hide-totals', '--totals',
        is_flag=True,
        help=_('Do not calculate and show totals in final row of output.'),
    ),
]


# ***
# *** [POST PROCESS] Adjust **kwargs.
# ***

def _postprocess_options_matching(kwargs):
    _postprocess_options_match_activities(kwargs)
    _postprocess_options_match_categories(kwargs)
    _postprocess_options_match_tags(kwargs)


def postprocess_options_normalize_search_args(kwargs, cmd_journal=False):
    _postprocess_options_matching(kwargs)
    _postprocess_options_grouping(kwargs, cmd_journal=cmd_journal)
    _postprocess_options_sort_cols(kwargs)
    _postprocess_options_sort_dirs(kwargs)
    _postprocess_options_results_show_hide(kwargs)
    postprocess_options_output_format_any_input(kwargs, cmd_journal=cmd_journal)
    _postprocess_options_output_filename(kwargs)
    _postprocess_options_sparkline(kwargs)
    _postprocess_options_search_term(kwargs)


# ***
# *** [ALL TOGETHER NOW] One @decorator is all you need.
# ***

# *** One @decorator for all your search command option needs.

def cmd_options_any_search_query(command='', item='', match=False, group=False):
    def _cmd_options_any_search_query():
        options = []
        append_cmd_options_filter_by_pk(options)
        append_cmd_options_filter_by_time(options)
        append_cmd_options_matching(options)
        append_cmd_options_group_by(options)
        append_cmd_options_results_sort_limit(options)
        append_cmd_options_results_column_choices(options)
        append_cmd_options_output_file(options)
        append_cmd_options_results_report_formats(options)
        append_cmd_options_results_report_totals(options)
        append_cmd_options_filter_by_search_terms(options)

        def _cmd_options_search_query(func):
            for option in reversed(options):
                func = option(func)
            return func

        return _cmd_options_search_query

    # +++

    def append_cmd_options_filter_by_pk(options):
        if command == 'export':
            return

        options.extend(_cmd_options_search_item_key)

    def append_cmd_options_filter_by_time(options):
        options.extend(_cmd_options_search_time_window(command))

    def append_cmd_options_filter_by_search_terms(options):
        options.extend(_cmd_options_search_search_term)

    # +++

    def append_cmd_options_matching(options):
        if not match:
            return

        # We could exclude the --activity item from the `dob list activity`
        # and `dob usage activity` commands, but by including it, we allow
        # the user access to a strict text match search, as opposed to using
        # search_term, which is a loose match. E.g., the difference between
        # `dob list activity -a "Must Match Fully"` as compared to
        # `dob list activity fully`, given Activity named "Must Match Fully".
        options.extend(_cmd_options_search_match_activities)
        options.extend(_cmd_options_search_match_categories)

        if item == 'fact':
            options.extend(_cmd_options_search_match_tags)
            options.extend(_cmd_options_search_broad_match)

    # +++

    def append_cmd_options_group_by(options):
        if not group:
            return

        if item != 'activity':
            options.extend(_cmd_options_results_group_activity)
        if item != 'category':
            options.extend(_cmd_options_results_group_category)
        if item != 'tags':
            options.extend(_cmd_options_results_group_tags)
        if item == 'fact':
            options.extend(_cmd_options_results_group_days)
        options.extend(_cmd_options_results_group(item))

    # +++

    def append_cmd_options_results_sort_limit(options):
        options.extend(_cmd_options_results_sort_order(item, command, group))
        options.extend(_cmd_options_search_limit_offset)

    # +++

    def append_cmd_options_results_column_choices(options):
        if command == 'export':
            return

        append_cmd_options_results_basic_usage_hide_show(options)
        append_cmd_options_results_fact_attrs_hide_show(options)

    def append_cmd_options_results_report_formats(options):
        if command == 'export':
            return

        append_cmd_options_output_format(options)
        append_cmd_options_table_type(options)
        append_cmd_options_results_max_width(options)

        append_cmd_options_format_factoid_output(options)
        append_cmd_options_format_sparkline_output(options)

        append_cmd_options_results_re_sort(options)

    def append_cmd_options_results_report_totals(options):
        if command == 'export':
            return

        if command == 'journal':
            options.extend(_cmd_options_output_show_totals)
        else:
            options.extend(_cmd_options_output_hide_totals)

    # +++

    def append_cmd_options_results_basic_usage_hide_show(options):
        # Search results report output column values hide/show options.
        if command == 'usage':
            options.extend(_cmd_options_results_hide_usage)
            options.extend(_cmd_options_results_hide_duration)
        elif command in ['list', 'journal']:
            options.extend(_cmd_options_results_show_usage)
            options.extend(_cmd_options_results_show_duration)
        elif command == 'export':
            pass  # Show neither.
        else:  # Impossible.
            raise False  # pragma: no cover

    def append_cmd_options_results_fact_attrs_hide_show(options):
        # Search results report output column values hide/show options.
        if item != 'fact':
            return

        options.extend(_cmd_options_results_hide_description)
        # (lb): I sorta wanna hide final_start and final_end by default.
        # - But then I'm afraid no one would find them.
        # - So instead, supply --column option to fine-tune the output.
        options.extend(_cmd_options_results_show_columns)

    def append_cmd_options_results_max_width(options):
        if command == 'export':
            return

        options.extend(_cmd_options_results_max_width)

    def append_cmd_options_results_re_sort(options):
        if item != 'fact':
            return

        # (lb): It'd be nice to squelch this option unless config['dev.catch_errors']
        #       but I guess that's what Click 'hidden' option feature is for.
        options.extend(_postprocess_options_results_options_sort_double_sort)

    # +++

    def append_cmd_options_output_file(options):
        for_export = (command == 'export')
        options.extend(_cmd_options_results_output_path(for_export=for_export))

    # +++

    def append_cmd_options_output_format(options):
        options.extend(_cmd_options_output_format_choices(command, item))

    def append_cmd_options_format_factoid_output(options):
        if item != 'fact':
            return

        options.extend(_cmd_options_output_factoids_hrule)

    def append_cmd_options_format_sparkline_output(options):
        if item != 'fact':
            return

        options.extend(_cmd_options_output_sparkline_format)

    def append_cmd_options_table_type(options):
        options.extend(_cmd_options_output_format_tabling())

    # +++

    return _cmd_options_any_search_query()




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

import sys
from inflector import English, Inflector

from nark.managers.query_terms import QueryTerms

from dob_bright.reports.render_results import render_results
from dob_bright.termio import (
    click_echo,
    dob_in_user_exit,
    dob_in_user_warning,
    highlight_value
)

from ..clickux.cmd_options_search import cmd_options_output_format_facts_only
from ..clickux.query_assist import error_exit_no_results

__all__ = (
    'list_facts',
)


def list_facts(
    controller,
    # - Save for controller, all parameters are CLI --options.
    # - The named keyword arguments listed first are used for output
    #   formatting and not for the query.
    # - Args: Select columns to output.
    show_usage=False,
    show_duration=False,
    hide_description=False,
    column=None,
    # - Args: Output formatter processor writer.
    output_format='csv',
    # - Args: Format- and Row-specific arguments.
    table_type='texttable',  # Applies when output_format == 'table'.
    max_width=-1,
    row_limit=None,
    factoid_rule='',
    # - Args: Output constraints.
    output_path=None,
    # - Args: Cell-specific arguments.
    spark_total=None,
    spark_width=None,
    spark_secs=None,
    # - Args: Total reporting arguments.
    show_totals=False,
    hide_totals=False,
    # - Developer controls.
    re_sort=False,
    # - Any unnamed arguments are used as search terms in the query.
    *args,
    # - All remaining keyword arguments correspond to nark.QueryTerms.
    **kwargs
):
    """
    Finds and lists facts according to specified filtering, sorting and display options.

    Writes to stdout, or to the file specified by ``output_path``.

    Arguments:
        Most of the arguments are documented elsewhere.

    Returns:
        None: If success.
    """
    format_restricted = output_format in cmd_options_output_format_facts_only()

    def _list_facts():
        # MAYBE: (lb): Verify QueryTerms output format options now, before searching.
        # - When I refactored query_terms, I moved validation later, so, e.g., you
        #   could specify an incorrect output_format, but it won't get caught until
        #   after find_facts returns. Which means the code does unnecessary processing,
        #   and the user has to wait a little longer until they're told they're wrong.
        qt = prepare_query_terms(*args, **kwargs)
        results = find_facts(controller, query_terms=qt)
        if not results:
            error_exit_no_results(_('facts'))
        n_total = len(results)
        n_written = display_results(results, qt, output_path)
        report_report_written(controller, output_path, n_total, n_written)

    # ***

    def prepare_query_terms(*args, **kwargs):
        qt = QueryTerms(*args, **kwargs)
        must_grouping_allowed(qt)
        qt.include_stats = should_include_stats(qt)
        qt.sort_cols = decide_sort_cols(qt)
        return qt

    def must_grouping_allowed(qt):
        if not qt.is_grouped or not format_restricted:
            return

        dob_in_user_exit(_(
            'ERROR: Specified format type does not support grouping results.'
        ))

    # Note that the two complementary commands, dob-list and dob-usage,
    # have complementary options, --show-usage and --hide-usage options,
    # and --show-duration and --hide-duration, that dictate if we need
    # the query to include the aggregate columns or not.
    def should_include_stats(qt):
        # If preparing a Factoid export, the FactoidWriter
        # expects a list of items, and not prepared tuples.
        if format_restricted:
            return False

        # If a group-by is specified, the report will show the aggregated names,
        # such as showing 'Activities' instead of 'Activity'. Include the extra
        # aggregate columns.
        if qt.is_grouped:
            return True

        # Not grouping results, so each result is a Fact, but check if the user
        # wants the 'Duration'. (It's an either-or with the SQL code, either you
        # retrieve just item columns, or you retrieve item columns and all extra
        # columns, whether or not you want them all.)
        # - The 'duration' is a simple julianday() - julianday() value, which
        #   the tabulate_results.prepare_duration may or may not use (it might
        #   call fact.delta instead) so really adding stats just to get the
        #   'Duration' is about as unnecessary as this comment.
        # - Ignore `show_usage` -- 'uses' obviously 1 because not qt.is_grouped.
        # - Check for 'sparkline' output, which is calculated from 'duration'.
        #   - Again, we could just do this during post-processing in tabulate_results.
        if (
            show_duration
            and (
                not column
                or 'duration' in column
                or 'sparkline' in column
            )
        ):
            return True

        return False

    def decide_sort_cols(qt):
        if qt.sort_cols is not None:
            return qt.sort_cols

        # If request includes the 'duration' column, might as well use it.
        if not qt.include_stats or not show_usage:
            # No 'duration' column, or user not asking to see it.
            return ('start',)  # The get_all default.
        return ('time',)  # Aka, sort-by 'duration'.

    # ***

    def find_facts(controller, **kwargs):
        """
        Search for one or more facts, given a set of search criteria and sort options.

        Args:
            query_terms: A QueryTerms object that defines the query.

            **kwargs: Alternatively, pass any or all of the QueryTerms attributes.

        Returns:
            A list of matching Facts, or matching (Fact, *statistics) tuples,
            depending on the QueryTerms.
        """
        try:
            return controller.facts.get_all(**kwargs)
        except Exception as err:
            # - NotImplementedError happens if db.engine != 'sqlite', because
            #   get_all uses SQLite-specific aggregate functions.
            # - ParserInvalidDatetimeException happens on bad since or until.
            dob_in_user_exit(str(err))

    # ***

    def display_results(results, qt, output_path):
        row_limit = suss_row_limit(qt)
        n_written = render_results(
            controller,
            results,
            query_terms=qt,
            show_usage=show_usage,
            show_duration=show_duration,
            hide_description=hide_description,
            custom_columns=column,
            output_format=output_format,
            table_type=table_type,
            max_width=max_width,
            row_limit=row_limit,
            factoid_rule=factoid_rule,
            output_path=output_path,
            spark_total=spark_total,
            spark_width=spark_width,
            spark_secs=spark_secs,
            show_totals=show_totals,
            hide_totals=hide_totals,
            re_sort=re_sort,
        )
        return n_written

    def suss_row_limit(qt):
        # Limit the number of rows dumped, unless user specified --limit,
        # or if not dumping to the terminal.
        _row_limit = row_limit
        # Note: Not caring about qt.limit here, as the query limit is
        # a separate concern from the row limit.
        if _row_limit is None and not output_path and sys.stdout.isatty():
            _row_limit = controller.config['term.row_limit']
        return _row_limit

    # ***

    def report_report_written(controller, output_path, n_total, n_written):
        if (
            not output_path
            or not sys.stdout.isatty()
            # (lb): I don't quite like this coupling, but it works:
            #       - Don't click_echo if carousel_active.
            or controller.ctx.command.name == 'edit'
        ):
            # If writ to stdout or pager, skip count and path report.
            return
        # Otherwise, path was formed from, e.g., "export.{format}", so display actual.

        if n_written < n_total:
            echo_warn_if_truncated(controller, n_written, n_total)

        click_echo(_(
            "Wrote {n_written} {facts} to {output_path}"
        ).format(
            n_written=highlight_value(n_written),
            facts=Inflector(English).conditional_plural(n_written, _('Fact')),
            output_path=highlight_value(output_path),
        ))

    def echo_warn_if_truncated(controller, n_results, n_rows):
        if n_results <= n_rows:
            return

        dob_in_user_warning(_(
            'Showed only {} of {} results. Use `-C term.row_limit=0` to see all results.'
        ).format(format(n_results, ','), format(n_rows, ',')))

    # ***

    _list_facts()




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from dob_bright.termio import attr, coloring

__all__ = (
    'help_header_format',
)


def help_header_format(text):
    return '{underlined}{text}{reset}{optolon}'.format(
        underlined=attr('underlined'),
        text=text,
        reset=attr('reset'),
        optolon=not coloring() and _(':') or '',
    )




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import os

from gettext import gettext as _

from nark.items.fact import Fact

# Profiling: load AppDirs: ~ 0.011 secs.
from dob_bright.config.app_dirs import AppDirs
from dob_bright.termio import ascii_art, attr, click_echo, fg, highlight_value

from .clickux.plugin_group import ClickPluginGroup

from . import get_version, __package_name__

__all__ = (
    'echo_app_details',
    'echo_app_environs',
    'echo_data_stats',
    'hamster_time',
)


def echo_app_details(controller, full=False):
    """List details about the runtime environment."""
    def _echo_app_details():
        echo_name_version()
        echo_config_path()
        echo_plugins_basepath()
        echo_logfile_path()
        echo_db_info()
        echo_app_dirs()

    def echo_name_version():
        click_echo(_(
            "You are running {appname} version {version}"
        ).format(
            appname=highlight_value(__package_name__),
            version=highlight_value(get_version()),
        ))

    def echo_config_path():
        click_echo(_(
            "Configuration file at: {}"
        ).format(
            highlight_value(controller.configurable.config_path),
        ))

    def echo_plugins_basepath():
        click_echo(_(
            "Plugins directory at: {}"
        ).format(
            # (lb): Such hack.
            highlight_value(ClickPluginGroup().plugins_basepath),
        ))

    def echo_logfile_path():
        click_echo(_(
            "Logfile stored at: {}"
        ).format(
            highlight_value(controller.config['log.filepath']),
        ))

    def echo_db_info():
        click_echo(get_db_info())

    def get_db_info():
        result = None
        # For now we do not need to check for various store option as we allow
        # only one anyway.
        result = get_sqlalchemy_info()
        return result

    def get_sqlalchemy_info():
        """"""
        def _get_sqlalchemy_info():
            engine = controller.config['db.engine']
            if engine == 'sqlite':
                return sqlalchemy_string_sqlite()
            else:
                return sqlalchemy_string_remote(engine)

        def sqlalchemy_string_sqlite():
            sqlalchemy_string = _(
                "Using {engine} on database: {db_path}"
                .format(
                    engine=highlight_value('sqlite'),
                    db_path=highlight_value(controller.config['db.path']),
                )
            )
            return sqlalchemy_string

        def sqlalchemy_string_remote(engine):
            port = controller.config['db.port']
            if port:
                port = ':{}'.format(port)
            sqlalchemy_string = _(
                "Using {engine} on database {db_name} at:"
                " {username}@{host}{port}".format(
                    engine=highlight_value(engine),
                    db_name=highlight_value(controller.config['db.name']),
                    username=highlight_value(controller.config['db.user']),
                    host=highlight_value(controller.config['db.host']),
                    port=highlight_value(port),
                )
            )
            return sqlalchemy_string

        return _get_sqlalchemy_info()

    def echo_app_dirs():
        if not full:
            return
        appdir_paths = existent_app_dirs(include_errs=True, highlight=True)
        for prop in sorted(appdir_paths.keys()):
            path = appdir_paths[prop]
            click_echo('AppDirs.{}: {}'.format(prop, highlight_value(path)))

    _echo_app_details()


def echo_app_environs(controller):
    """List details about runtime in source'able name=value format."""
    environs = {}

    def _echo_app_environs():
        environs_add_all()
        environs_echo()

    def environs_add_all():
        environs_add_appname_ver()
        environs_add_config_path()
        environs_add_plugins_path()
        environs_add_log_path()
        environs_add_db_url()
        environs_add_db_path()
        environs_add_user_app_dirs()

    def environs_echo():
        for key in sorted(environs.keys()):
            val = environs[key]
            click_echo('DOB_{}="{}"'.format(key.upper(), val))

    def environs_add_appname_ver():
        environs['appname'] = __package_name__
        environs['version'] = get_version()

    def environs_add_config_path():
        environs['conf'] = controller.configurable.config_path

    def environs_add_plugins_path():
        environs['plugins'] = ClickPluginGroup().plugins_basepath

    def environs_add_log_path():
        # environs['log_filepath'] = controller.config['log.filepath']
        environs['CONFIG_LOG_FILEPATH'] = controller.config['log.filepath']

    def environs_add_db_url():
        environs['db_url'] = controller.store.db_url

    def environs_add_db_path():
        if controller.config['db.path']:
            # environs['db_path'] = controller.config['db.path']
            environs['CONFIG_DB_PATH'] = controller.config['db.path']

    def environs_add_user_app_dirs():
        for prop, path in existent_app_dirs().items():
            environs[prop] = path

    return _echo_app_environs()


def existent_app_dirs(include_errs=False, highlight=False):
    """"""
    def _existent_app_dirs():
        was_create = appdirs_disable_create()
        prop_paths = build_prop_paths()
        appdirs_restore_create(was_create)
        return prop_paths

    def appdirs_disable_create():
        # (lb): Our AppDirs defaults to trying to create directories. Tell it not to.
        #   NOTE: I consider this side effect a bug. Paths should be lazy-created only
        #   when a file stored therein is opened for writing.
        was_create = AppDirs.create
        AppDirs.create = False
        return was_create

    def appdirs_restore_create(was_create):
        AppDirs.create = was_create

    def build_prop_paths():
        actual_prop_paths = {}

        # CONFIRM: (lb): On virtualenv install, no site_* dirs.
        #   Will this not be the case for site-wide, non-virtenv installs?
        for prop in [
            'user_data_dir',
            'site_data_dir',
            'user_config_dir',
            'site_config_dir',
            'user_cache_dir',
            'user_log_dir',
        ]:
            path = getattr(AppDirs, prop)
            path = check_exists(path)
            if path is not None:
                actual_prop_paths[prop] = path

        return actual_prop_paths

    def check_exists(path):
        if not os.path.exists(path):
            if include_errs:
                dne = _(' [dne]')
                if highlight:
                    dne = _('{}{}{}').format(fg('red_3b'), dne, attr('reset'))
                path += dne
            else:
                path = None
        return path

    return _existent_app_dirs()


def echo_data_stats(controller):
    def _echo_data_stats():
        echo_counts()
        echo_facts_interesting()

    def echo_counts():
        # MAYBE: Add filtering, like activity, category, search_term, after, until, etc.
        num_activities = controller.activities.get_all(count_results=True)
        num_categories = controller.categories.get_all(count_results=True)
        num_facts = controller.facts.get_all(count_results=True)
        num_tags = controller.tags.get_all(count_results=True)
        # FIXME: Calculate length of Hamstering, i.e., delta first and last facts.
        # FIXME: Other stats?
        click_echo(_("No. of      Facts: {}").format(highlight_value(num_facts)))
        click_echo(_("No. of  Tag names: {}").format(highlight_value(num_tags)))
        click_echo(_("No. of Activities: {}").format(highlight_value(num_activities)))
        click_echo(_("No. of Categories: {}").format(highlight_value(num_categories)))

    def echo_facts_interesting():
        first_facts = controller.facts.get_all(
            sort_cols=('start',), sort_orders=('asc',), limit=1,
        )
        final_facts = controller.facts.get_all(
            sort_cols=('start',), sort_orders=('desc',), limit=1,
        )
        assert len(first_facts) <= 1
        assert len(final_facts) <= 1
        if not first_facts:
            assert not final_facts
            return
        first_fact = first_facts.pop()
        final_fact = final_facts.pop()
        time_0 = first_fact.start
        time_n = final_fact.end or controller.now
        spanner = Fact(activity=None, start=time_0, end=time_n)
        elapsed = spanner.format_delta(style='')
        click_echo(_("Lifetime of Facts: {}").format(highlight_value(elapsed)))

    _echo_data_stats()


def hamster_time(posits=[]):
    if posits:
        arts = ascii_art.fetch_asciis(posits)
    else:
        arts = [ascii_art.randomster()]

    for one_art_please in arts:
        click_echo(one_art_please)




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Module to provide u18n friendly help text strings for our CLI commands."""

from gettext import gettext as _

from .help_strings import common_format


# ***
# *** [CREATE-FACT] Commands help.
# ***

# FIXME: Thoughts on format.
#        If always an '@' you could know if activity specified,
#           and not start of description.
#        OR: You could require that an Activity be specified!
#          (And you don't need category if it can be inferred; or if you
#           use a default.)
#          (You could say that you could have an Activity default, but I
#          don't want to encourage not categorizing your Facts, so require
#          an Activity -- if anything, a lazy user could either make a Bash
#          alias, or make a catch-all Activity, e.g.,
#           `hamster at now null I did something.`
#
#        For #tags, you could require that if multiple words, they're wrapped
#           in quotes.
#        Otherwise, if just single words, then you don't need that silly comma
#           in the factoid!
#        And also, you could easily mistake, say,
#          `hamster at now null #hashtag I went to market, but didn't buy anything`
#        for the hashtag: "hashtag I went to the market", which is what
#           the code currently does!
#
#        For timestamp, require hours and minutes; but allow any obvious delimiter.

# NOTE: In Click, \b prevents re-wrapping blocks of text using terminal width.
def ADD_FACT_COMMON(ctx):
    # Throw out a hint to use the pager, if not being used,
    # because this help document is dense.
    first_line = 'Help on Adding single Facts from the command line'
    if not ctx.find_root().params['pager']:
        first_line = _(
            'Show Add Fact help'
            ' (or try {codehi}{rawname} -P add -h{reset} for paged help)'
        ).format(**common_format())

    # - Tricky, tricky: Click truncates first line at first period '.',
    # but if no dot found Click shows ellipses '...'. Show use a '.'.
    # - Also, each paragraph is reformatted to a common indentation
    # (the same amount of leading whitespace is removed from each line
    # in the string below, and then the lines are re-indented), but a
    # line above with just \b will skip re-indenting that paragraph.
    # After a \b, the first space of indent is converted to 2. So the
    # spacing after a \b might appear off by one.
    _help = _(
        """
        {first_line}.

        {underlined}Overview{reset}

        In {appname}, there are two easy ways to start and stop Facts:

        - You can use the interactive editor, by running {codehi}{rawname} edit{reset}.

        - You can use any of the Add Fact commands, documented below.

        {underlined}Concepts{reset}

        In {appname}, a {bold}Fact{reset} is essentially
        a document with a start time and an end time.
        It's a single timesheet entry.

        You can use {appname} to track time for your job,
        or maybe you use {appname} as a life hacking tool
        and you record every droll moment.

        However you use it, each Fact represents a block
        of time, with a start time, an end time (possibly
        left blank), your comments, and your tags.

        You can add Facts after you complete them,
        if you specify a start time and an end time.
        These Facts are considered {bold}complete{reset}.

        You can also {italic}{appname}{reset} in real time, if you
        specify a start time but not an end time.
        A Fact without an end time is considered the
        {bold}active{reset} Fact.

        To avoid chaos, you can only {bold}start{reset} one Fact at
        a time. So there will only be one active Fact at a time
        (or none).

        Once started, a Fact is {italic}active{reset} until you
        {bold}stop{reset} it, by setting its end time.

        {underlined}Why Are There So Many Add Fact Commands?{reset}

        There are a lot of commands that can be used to start a Fact.

        Each Add Fact command may perform either or both of these actions:

        - Stop the presently active Fact, if one was previously started.

        - Create and start a new Fact, making it the active Fact.

        Each of the different Add Fact commands performs a slightly
        different variation of those two actions, including how they
        set the start and end times.

        The variety of Add Fact commands should help you be more efficient.
        It should hopefully make adding Facts a little more fun, too, by
        making it feel more natural to interact with {appname}.

        {underlined}General Format{reset}

        The general Add Fact command format is:

         \b
         {codehi}{rawname} [COMMAND] [START-TIME] [to] [END-TIME] \\{reset}
             {codehi}[<activity>[@<category>]] [@<tag>...] [<description>]{reset}

        When you start or stop a Fact, you might specify one or two
        time values, which can be expressed using a timestamp
        (e.g., "2019-01-23"), or using more natural language
        (e.g., "5 mins. ago").

        - See {wordhi}{underlined}Time Format{reset}, below, for more help.

        When you start a new Fact, you can optionally specify
        an Activity, a Category, Tags, and a Description.

        - See {wordhi}{underlined}Meta Formats{reset}, below, for more help.

        Not all commands accept the start or end time arguments.

        Here's a look at which commands expect which time arguments.

         \b
         Command   Start   End   Example
         -------   -----   ---   -------
         {codehi}now{reset}          No    No   {codehi}{rawname} now Baking cookies.{reset}
         {codehi}at{reset}          Yes    No   {codehi}{rawname} at -10m: Eating cookies.{reset}
         {codehi}from{reset}        Yes   Yes   {codehi}{rawname} from 5:00 PM to 6:00 PM Workin' out.{reset}
         {codehi}to{reset}           No   Yes   {codehi}{rawname} to 5 mins. ago: Finishing task.{reset}
         {codehi}then{reset}        Opt    No   {codehi}{rawname} then I started a new project.{reset}
         {codehi}still{reset}       Opt    No   {codehi}{rawname} still Built Part 1, now Part 2.{reset}
         {codehi}after{reset}       Opt    No   {codehi}{rawname} after +5m: Left for home.{reset}

        Some hints:

        - The {codehi}from{reset} command expects 'to' between the start and end times.

        - If you do not specify a time when calling
        {codehi}then{reset}, {codehi}still{reset}, or {codehi}after{reset},
        the command will assume the time is "now" (the current machine time).
        The {codehi}now{reset} command always assumes the time is "now".

        - The {codehi}to{reset} command does not start a new Fact.
        It stops the active Fact, or if there is no active Fact,
        it starts a new Fact at the time when the last Fact ends
        and ends the new Fact at the time specified.

        - The {codehi}from{reset} command also does not start
        a new Fact. It creates a new Fact, but completes it
        (sets both the start time and the end time).

        - The {codehi}now{reset} and {codehi}at{reset} commands only
        start a new Fact. Alternatively, the
        {codehi}then{reset}, {codehi}still{reset}, and {codehi}after{reset}
        commands also start a new Fact but also end the presently active Fact.

        {underlined}Continuous Tracking{reset}

        Be aware that stopping the active Fact might lead you to inadvertently
        create time gaps in the data.

        For instance, suppose you call {codehi}{rawname} now{reset} to
        start a new Fact, and then later you {codehi}{rawname} stop{reset}
        to complete the Fact.

        After that, if you call {codehi}{rawname} now{reset} to start
        a second Fact, there will be a gap in the
        data from between the time you called {codehi}stop{reset} and the
        time you executed the second {codehi}now{reset}.

        A better method, which creates a seamless Fact lineage,
        is to never use {codehi}{rawname} stop{reset},
        and to always maintain an active Fact,
        or to always start new Facts using the end time of the latest completed Fact.

        A simple example is always using {codehi}now{reset}.
        If you run {codehi}{rawname} now{reset} once, when you first
        install {appname} and begin dobbing, you could just call
        {codehi}{rawname} now{reset} again whenever you want to create
        and start a new Fact.

        The {codehi}now{reset} command will close whatever Fact is
        active before starting a new active Fact, ending the old
        Fact with the same time used to start the new Fact.

        As another example, if you switched what you were working
        on but forgot to dob it, you could specify the time in the
        past when the old Fact ends and the new Fact starts, e.g.,

         \b
         {codehi}{rawname} at -1h: Built a thing!{reset}

        As yet another example, if you already stopped a Fact,
        you could create a new Fact that starts when the
        stopped Fact ends
        by using the {codehi}{rawname} then{reset} command,
        which will create a new Fact and set the start time to the end
        time of the most recently completed Fact. E.g.,

         \b
         {codehi}{rawname} then Built another thing!{reset}

        As a general rule, if an Add Fact command both starts a new Fact
        and stops an old Fact, the stopped Fact will always end exactly
        when the new Fact starts. This maintains a seamless, linear Fact
        database. If you want to create gaps in the data, you can
        use {codehi}stop{reset},
        or you can use the {codehi}to{reset}
        command with a relative negative time.

        {underlined}How to Use Activity, Category, and Tags{reset}

        Each Fact can be assigned to one Activity, and each Activity
        belongs to one Category. (So you could have two Activities with
        the same name, but in different Categories.)

        You can assign multiple Activities to the same Category to group
        Activities together.

        The Activity and Category for a Fact, colloquially called an
        {bold}act@gory{reset}, is useful for generating reports.

        Because each Fact can only be assigned one Activity and one Category
        combination, {rawname} can easily group Facts based on their
        Activity and Category to generate reports and to calculate statistics.

        You can also filter reports and search Facts using act@gories.

        Each Fact can also be assigned as many tags as you like.

        Tags are also useful for searching Facts,
        and Tags are useful for generating reports,
        but they're more work than relying on the {italic}act@gory{reset},
        because you need to specify the list of tags on which to report.
        Also, because you can apply more than one tag to a Fact,
        a Fact might unintentionally be used in multiple parts
        of a report or statistics calculation.

        {underlined}Time Format{reset}

        The start and end time arguments can each be either a timestamp,
        or a friendly string.

        - A timestamp can be as simple as {codehi}HH:MM{reset}
        to specify hours and minutes of the present day.

        - A timestamp can also include a
        {codehi}YYYY-MM-DD{reset}-formatted
        year, month, and day.

        - A friendly string is a human-readable phrase
        such as "now", "yesterday", "2 hours ago", etc.

        The Time Format must conform to one of the following rules:

         \b
         Format            Description
         ------            -----------
         {codehi}HH:MM{reset}             Clock time of present day, in 24 H military time.
         {codehi}HH:MM [AM|PM]{reset}     Clock time of present day, in 12 H civilian time.
         {codehi}YYYY-MM-DD HH:MM [AM|PM]{reset}
                           Year, month, day, and clock time.
         {codehi}-MM[m|h]{reset}          Minutes or Hours before "now", or nearest later Fact.
         {codehi}+MM[m|h]{reset}          Minutes or Hours after start or end nearest to "now".
         <friendly-time>   "1 week ago", "45 mins ago", "yesterday @ 3 PM", etc.

        - To be clear, {codehi}[AM|PM]{reset} means use either
        {codehi}AM{reset} or {codehi}PM{reset}.

        {underlined}Relative Time{reset}

        To really confuse you, dear reader, if you're not already overwhelmed,
        note that a positive or negative relative time duration can be used
        to specify the time.

        A negative duration generally resolves relative to "now", e.g.,
        when using an add Fact command, using "-1h" resolves to one hour
        before the current machine time.

        A positive duration used when adding a new Fact, alternatively,
        resolves relative to the start or end time of another Fact in
        the database. When using an add Fact command, a positive duration
        is computed relative to either the start time of the active Fact,
        or to the end time of the most recently completed Fact.

        For instance, if there is no active Fact, you could start
        one relative to when the latest Fact was completed, e.g., to
        start a new Fact 1 hour after the latest Fact was ended, run

         \b
         {codehi}{rawname} at +1h: Back at it.{reset}

        Alternatively, to reference the start time of the active
        Fact, say, to complete the active Fact and make it exactly
        one hour long, and to also append text to the description,
        run:

         \b
         {codehi}{rawname} to +1h: Finished task.{reset}

        Without getting too into the weeds, we'll also let you know
        that the {codehi}import{reset} command understands
        {italic}compounding{reset} relative times.

        E.g., suppose you were not able to dob until the day ended,
        and you wanted to record 3 tasks. The first task took 1 hour,
        the second task took 2 hours, and the third task took 3 hours.
        You could create an input stream with the three entries
        and then import that:

         \b
         {codehi}cat <<EOF |{reset}
         {codehi}{rawname} at 09:00 AM: Working on Task 1.{reset}

         \b
         {codehi}{rawname} at +1h: Working on Task 2.{reset}

         \b
         {codehi}{rawname} from +2h to +3h: Working on Task 3.{reset}
         {codehi}EOF{reset}
         {codehi}dob import{reset}

        Note that the {codehi}+3h{reset} end time of the third Fact
        will not be computed until after the start time ({codehi}+2h{reset})
        is computed, and that the length of the third Fact is 3 hours, not 1
        (i.e., it doesn't start 2 hours after the Fact before it, and then
        end 3 hours after the Fact before it; it starts 2 hours after the
        Fact before it, and then it ends 3 hours after it starts!).

        In general, these two rules apply to positive relative times:

        - For an end time, a positive relative time is relative to
        the same Fact's start time.

        - For a start time, a positive relative time is relative
        to the end time of the Fact before it.

        A negative relative time, on the other hand, will have the
        opposite properties:

        - If used for a start time, a negative relative time will
        be relative that that Fact's end time, or to "now", if there
        is no end time. (Except during import, than a negative relative
        time might be relative to the Fact that follows it!)

        - If used for an end time, a negative relative time is
        relative to "now" (except, again, for import, and then it's
        relative to the Fact after it).

        {underlined}Meta Formats{reset}

        {underlined}Act@Gory{reset}

        The Activity name and Category name are combined into one
        argument when specified on the command line. Use the at
        symbol, {codehi}@{reset}, to separate the two values (so
        you can use spaces in your Activity and Category names,
        but not the @ symbol!).

        HINT: Each Category contains one or more Activities, and each
        Activity belongs to one or more Categories, but each Activity
        will only be associated with any one Category at most once.

        For example, you could use the Activity to track different
        projects at work. Run an Add Fact command to start the first
        project using its name as the Activity name,

         \b
         {codehi}{rawname} now "Project-XYZ@My Job": Working on it.{reset}

        and then later run another command to start a new Fact with
        the other project name as the Activity name,

         \b
         {codehi}{rawname} now "Project-ABC@My Job": All Hands Meeting.{reset}

        In dob, the act@gory and tags are somewhat interchangeable.

        For instance, considering the previous example,
        you could track the same two projects using tags,
        starting on the first project with:

         \b
         {codehi}{rawname} now "Work@My Job" @Project-XYZ: Working on it.{reset}

        and then starting the second project with:

         \b
         {codehi}{rawname} now "Work@My Job" @Project-ABC: Working on it.{reset}

        The one drawback to using tags is that you might have to
        specify a few more options to the report and search commands
        to achieve what you can do less verbosely with Activities and
        Categories.

        However, if you work on lots of different projects that you'd
        like to track, by using tags you can keep a lot of noise from
        infesting your activities and categories as projects come and go.

        One way to decide when to use an Activity to track something
        and when to use a Tag is to determine how enduring is the
        think you are tracking.

        If you are tracking a particular project that you'll complete
        in the next weeks or months, consider using a tag.

        If you are tracking an event that is constantly recurring,
        such as time spent commuting to and from work, consider
        using an Activity (and Category, i.e., an act@gory).

        {underlined}Tags{reset}

        You can specify zero or more Tags for each Fact.

        You must delimit each tag argument. The tag delimiter will
        not be included as part of the tag.

        It's easiest to use the {codehi}@{reset} symbol to denote tags,
        rather than using {codehi}#{reset}, like you'd see in social media.
        This is because you'd probably need to escape the pound symbol to use it.
        That is, because Bash and other shells interpret an octothorpe
        as the start of a comment, you have to use quotes around it ({codehi}"#"{reset})
        or escape it ({codehi}\#{reset}) to use it.

        E.g., these are all acceptable tags:

         \b
         {codehi}@tag1 @"tag2 too" "#tag3" @give-it-up-for-tag4 \#tag5 '#'tag6{reset}

        Also be aware that you need to use quotes if a tag contains whitespace.
        E.g., these are all acceptable ways to specify tags:

         \b
         {codehi}@foo-bar{reset}, {codehi}"@foo bar"{reset} and {codehi}@"foo bar"{reset}

        But without the quotes, e.g.,

         \b
         {codehi}@foo bar{reset}

        the parser would identify a tag named "foo" for the Fact,
        and the second part, "bar", would be seen as the description.

        {underlined}Description{reset}

        A description of the Fact for the particular Activity being performed
        at the indicated time. Which is to say, journal away! This is your
        worklog entry, or your diary entry, or whatever it is you want to
        call it. Record what you're doing, or what you're feeling, dig it?

        {underlined}Examples{reset}

        - Start your day by starting a Fact under the "Waking up"
        Activity in the "Personal" category. Suppose you got up
        at 8am and got to your machine 20 minutes later, you could
        {rawname}:

         \b
         {codehi}{rawname} at 08:00 Waking up@Personal Woke up. \\{reset}
         {codehi}  Slept pretty well, except for the spiders again.{reset}

        - Get ready for work, and then use the 'to' command and an empty '@'
        symbol to indicate the same Activity being closed.

         \b
         {codehi}{rawname} to 08:45 @ Getting ready for work.{reset}

        - Commute to work, then dob in the missing gap.
        Use a tag to remember what mode you used to travel.

         \b
         {codehi}{rawname} until 09:15 Commuting \#bike{reset}

        - Start working on a ticket.

         \b
         {codehi}{rawname} on PROJ-1234@Work Fixing all the bugs!{reset}

        - Go to lunch, but get back just in time to run to a meeting.

         \b
         {codehi}{rawname} from -30now @MEET-4567 @standup Barely made it to the meeting, oops!{reset}

         \b
         {codehi}{rawname} now @MEET-4567 @standup Barely made it to the meeting, oops!{reset}

        - Remember that you forgot to clock out for lunch and do that now.

         \b
         {codehi}{rawname} from 1130 to noon Eating@Personal #gram-it! Salad and fries.{reset}

        - Start working on another ticket and then remember to clock out of
        that earlier meeting. (HINT: `start` is an alias for `at`)

         \b
         {codehi}{rawname} start 30 mins ago '#PROJ-9876' @ Refactoring this mess.{reset}

        - All done for the day! Start a "Sleep" fact under "Personal".

         \b
         {codehi}{rawname} at 11:59 PM Sleep@Personal Time for bed!{reset}

        {underlined}Quick Reference{reset}

        Here's a quick look at which commands accept which time arguments,
        and how they behave.

         \b
         Command   Start   End   Stops    Default
                               Active?      Start
         -------   -----   ---   -----    ---------------------------
         {codehi}after{reset}       Opt    No     Y??    Last end time.
         {codehi}at{reset}          Yes    No     Yes    You must specify start.
         {codehi}from{reset}        Yes   Yes      Y?    You must specify both.
         {codehi}now{reset}          No    No     Yes    Always "now" time, when command run.
         {codehi}still{reset}       Opt    No     Yes    Last end time, or now; copies metadata.
         {codehi}then{reset}        Opt    No     Yes    Last end time, or now.
         {codehi}to{reset}           No   Yes     Yes    You must specify end time.
                                          If no active, new Fact starts at last end.

        """.format(first_line=first_line, **common_format())  # noqa: E501
    )
    return _help




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""A lite wrapper around the dob-bright Controller."""

from dob_bright.controller import Controller
from dob_bright.styling.apply_styles import pre_apply_style_conf

__all__ = (
    'Controller',
)


class DobController(Controller):
    """
    A custom controller that ensures the style config is ready when needed.
    """

    def __init__(self, *args, **kwargs):
        super(DobController, self).__init__(*args, **kwargs)
        self.applied_style_conf = False

    def setup_logging(self, *args, **kwargs):
        self.pre_apply_style_conf()
        return super(DobController, self).setup_logging(*args, **kwargs)

    def standup_store(self, *args, **kwargs):
        self.pre_apply_style_conf()
        return super(DobController, self).standup_store(*args, **kwargs)

    def pre_apply_style_conf(self):
        if self.applied_style_conf:
            return
        # Trigger the cache mechanism for the style conf.
        pre_apply_style_conf(self)
        self.applied_style_conf = True




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import os

import fauxfactory
import pytest

import nark
from nark.tests.conftest import *  # noqa: F401, F403
from nark.tests.backends.sqlalchemy.conftest import *  # noqa: F401, F403

from dob_bright.crud.fact_dressed import FactDressed
from dob_bright.reports.tabulate_results import report_table_columns

from dob.cmds_list.fact import list_facts

from .. import truncate_to_whole_seconds


# ***

class TestCmdsListFactListFacts_Simple(object):
    """"""

    def test_list_facts_empty_store(self, controller_with_logging, capsys):
        """..."""
        controller = controller_with_logging
        with pytest.raises(SystemExit):
            list_facts(
                controller,
            )
            assert False  # Unreachable.
        out, err = capsys.readouterr()
        # See: error_exit_no_results.
        expect = 'No facts were found for the specified query.'
        assert err.startswith(expect)

    # ***

    def test_list_facts_basic_store(
        self,
        five_report_facts_ctl,
        capsys,
    ):
        controller = five_report_facts_ctl
        list_facts(controller)
        out, err = capsys.readouterr()
        assert not err
        # We could verify the report looks like how we expect, but that seems
        # better suited for using a snapshots feature. It would be painful to
        # maintain here, especially as the report formats evolve.
        assert out


# ***

# Mostly coverage of
#   dob/cmds_list/fact.py
# but also provides coverage of
#   dob/facts_format/factoid.py

class TestCmdsListFactListFacts_PresentationArguments(object):

    # We're just going for coverage and not crashing, so not checking test outputs.

    # ***

    @pytest.mark.parametrize(
        ('include_stats', 'show_usage', 'show_duration', 'hide_description',), (
            (True, True, True, True),
        )
    )
    def test_list_facts_show_show_hide(
        self,
        five_report_facts_ctl,
        include_stats,
        show_usage,
        show_duration,
        hide_description,
    ):
        controller = five_report_facts_ctl
        list_facts(
            controller,
            include_stats=include_stats,
            show_usage=show_usage,
            show_duration=show_duration,
            hide_description=hide_description,
        )

    # ***

    def test_list_facts_custom_columns_all_columns(
        self,
        five_report_facts_ctl,
    ):
        controller = five_report_facts_ctl
        # See also: FACT_TABLE_HEADERS.
        custom_columns = report_table_columns()
        list_facts(controller, column=custom_columns)

    # ***

    # For reference:
    #
    #   REPORT_COLUMNS_DEFAULT = [
    #       'key',
    #       'start',
    #       'end',
    #       'activity',
    #       'category',
    #       'tags',
    #       'description',
    #       'duration',
    #   ]

    def test_list_facts_custom_columns_a_few_columns_group_activity(
        self,
        five_report_facts_ctl,
    ):
        controller = five_report_facts_ctl
        custom_columns = ['activity', 'duration']
        list_facts(controller, column=custom_columns, group_activity=True)

    # ***

    @pytest.mark.parametrize(
        ('output_format', 'table_type',), (
            ('table', 'normal'),
            ('table', 'rst'),
            ('factoid', '*ignored*'),
            ('journal', '*ignored*'),
        )
    )
    def test_list_facts_format_permutations(
        self,
        five_report_facts_ctl,
        output_format,
        table_type,
    ):
        controller = five_report_facts_ctl
        # Ensure that fact.friendly_str() is the FactDressed version
        # (because Factoid format uses 'colorful' argument).
        controller.store.fact_cls = FactDressed
        list_facts(
            controller,
            output_format=output_format,
            table_type=table_type,
        )

    # ***

    @pytest.mark.parametrize(
        ('spark_total', 'spark_width', 'spark_secs',), (
            ('', None, None),
            ('max', None, None),
            ('net', None, None),
            (100, None, None),
            (None, 24, None),
            (None, None, 3600),
        )
    )
    def test_list_facts_format_jounrnal_spark_args(
        self,
        five_report_facts_ctl,
        spark_total,
        spark_width,
        spark_secs,
    ):
        controller = five_report_facts_ctl
        list_facts(
            controller,
            output_format='journal',
            spark_total=spark_total,
            spark_width=spark_width,
            spark_secs=spark_secs,
        )

    # ***

    @pytest.mark.parametrize(
        ('output_format',), (
            ('table',),
            ('factoid',),
        )
    )
    def test_list_facts_max_width(
        self,
        five_report_facts_ctl,
        output_format,
    ):
        controller = five_report_facts_ctl
        list_facts(
            controller,
            output_format=output_format,
            max_width=80,
        )

    # ***

    def test_list_facts_sort_cols(
        self,
        five_report_facts_ctl,
    ):
        controller = five_report_facts_ctl
        sort_cols = ('activity', 'start')
        list_facts(controller, sort_cols=sort_cols)

    # ***


# ***

class TestCmdsListFactListFacts_FactoidPermutations(object):

    # We're just going for coverage and not crashing, so not checking test outputs.

    # ***

    def test_list_facts_output_format_factoid_factoid_rule(
        self,
        five_report_facts_ctl,
    ):
        controller = five_report_facts_ctl
        factoid_rule = '++rule-this++'
        list_facts(
            controller,
            output_format='factoid',
            factoid_rule=factoid_rule,
        )

    # ***

    def test_list_facts_output_format_factoid_output_path(
        self,
        five_report_facts_ctl,
        mocker,
    ):
        controller = five_report_facts_ctl
        output_path = mocker.MagicMock()
        list_facts(
            controller,
            output_format='factoid',
            output_path=output_path,
        )
        assert output_path.write.called

    # ***

    @pytest.mark.parametrize(
        ('output_format', 'row_limit'), (
            ('table', 2),
            ('factoid', 2),
        )
    )
    def test_list_facts_output_format_factoid_row_limit(
        self,
        five_report_facts_ctl,
        output_format,
        row_limit,
    ):
        controller = five_report_facts_ctl
        list_facts(
            controller,
            output_format=output_format,
            row_limit=row_limit,
        )

    # ***

    def test_list_facts_output_format_factoid_max_width(
        self,
        five_report_facts_ctl,
    ):
        controller = five_report_facts_ctl
        max_width = 40
        list_facts(
            controller,
            output_format='factoid',
            max_width=max_width,
        )

    # ***

    def test_list_facts_output_format_factoid_and_chop_isatty_uses_term_size(
        self,
        five_report_facts_ctl,
        mocker,
    ):
        controller = five_report_facts_ctl
        isatty = mocker.patch('sys.stdout.isatty', return_value=True)
        get_ts = mocker.patch(
            'click_hotoffthehamster.get_terminal_size', return_value=(80, 24),
        )
        list_facts(
            controller,
            output_format='table',
            max_width=None,
        )
        assert isatty.called
        assert get_ts.called

    # ***

    def test_list_facts_output_format_factoid_grouping_not_allowed(
        self,
        five_report_facts_ctl,
    ):
        controller = five_report_facts_ctl
        with pytest.raises(SystemExit):
            list_facts(
                controller,
                output_format='factoid',
                group_days=True,
            )
            assert False  # Unreachable.


# ***

class TestCmdsListFactListFacts_OutputFormats(object):
    """Unittests related to data export."""
    @pytest.mark.parametrize('output_format', ['soap', fauxfactory.gen_latin1()])
    def test_invalid_format(self, five_report_facts_ctl, output_format, mocker):
        """Make sure that passing an invalid format exits prematurely."""
        controller = five_report_facts_ctl
        with pytest.raises(Exception):
            list_facts(controller, output_format=output_format)

    def test_csv(self, five_report_facts_ctl, mocker):
        """Make sure that a valid format returns the appropriate writer class."""
        mocker.patch.object(nark.reports.csv_writer.CSVWriter, 'write_facts')
        controller = five_report_facts_ctl
        list_facts(controller, output_format='csv')
        assert nark.reports.csv_writer.CSVWriter.write_facts.called

    def test_tsv(self, five_report_facts_ctl, mocker):
        """Make sure that a valid format returns the appropriate writer class."""
        mocker.patch.object(nark.reports.tsv_writer.TSVWriter, 'write_facts')
        controller = five_report_facts_ctl
        list_facts(controller, output_format='tsv')
        assert nark.reports.tsv_writer.TSVWriter.write_facts.called

    def test_ical(self, five_report_facts_ctl, mocker):
        """Make sure that a valid format returns the appropriate writer class."""
        mocker.patch.object(nark.reports.ical_writer.ICALWriter, 'write_facts')
        controller = five_report_facts_ctl
        list_facts(controller, output_format='ical')
        assert nark.reports.ical_writer.ICALWriter.write_facts.called

    def test_xml(self, five_report_facts_ctl, mocker):
        """Ensure passing 'xml' as format returns appropriate writer class."""
        mocker.patch.object(nark.reports.xml_writer.XMLWriter, 'write_facts')
        controller = five_report_facts_ctl
        list_facts(controller, output_format='xml')
        assert nark.reports.xml_writer.XMLWriter.write_facts.called

    def test_with_since(self, controller, mocker):
        """Make sure that passing a end date is passed to the fact gathering method."""
        # (lb): Not sure utility of this test. It was from hamster-lib, so I
        # probably refactored away any utility.
        mocker.patch.object(controller.facts, 'gather')
        since = fauxfactory.gen_datetime()
        # Get rid of fractions of a second.
        since = truncate_to_whole_seconds(since)
        list_facts(
            controller,
            output_format='csv',
            since=since.strftime('%Y-%m-%d %H:%M'),
        )
        args, kwargs = controller.facts.gather.call_args
        query_terms = args[0]
        assert query_terms.since == since

    def test_with_until(self, controller, mocker):
        """Make sure that passing a until date is passed to the fact gathering method."""
        mocker.patch.object(controller.facts, 'gather')
        until = fauxfactory.gen_datetime()
        # Get rid of fractions of a second.
        until = truncate_to_whole_seconds(until)
        list_facts(
            controller,
            output_format='csv',
            until=until.strftime('%Y-%m-%d %H:%M'),
        )
        args, kwargs = controller.facts.gather.call_args
        query_terms = args[0]
        assert query_terms.until == until

    def test_with_filename(self, five_report_facts_ctl, tmpdir, mocker):
        """Make sure that a valid format returns the appropriate writer class."""
        path = os.path.join(tmpdir.ensure_dir('export').strpath, 'export.csv')
        mocker.patch.object(nark.reports.csv_writer.CSVWriter, 'write_facts')
        controller = five_report_facts_ctl
        list_facts(controller, output_format='csv', output_path=path)
        assert nark.reports.csv_writer.CSVWriter.write_facts.called




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from click_hotoffthehamster import ClickException

import pytest

from dob.facts.cancel_fact import cancel_fact


class TestCancel(object):
    """Unit tests related to cancelation of an ongoing fact."""

    def test_cancel_existing_ongoing_fact(
        self, ongoing_fact, controller_with_logging, mocker, capsys,
    ):
        """Test cancelation in case there is an ongoing fact."""
        controller = controller_with_logging
        current_fact = mocker.MagicMock(return_value=ongoing_fact)
        controller.facts.cancel_current_fact = current_fact
        cancel_fact(controller)
        out, err = capsys.readouterr()
        assert controller.facts.cancel_current_fact.called
        assert out.startswith('Cancelled: ')

    def test_cancel_no_existing_ongoing_fact(self, controller_with_logging, capsys):
        """Test cancelation in case there is no actual ongoing fact."""
        with pytest.raises(ClickException):
            cancel_fact(controller_with_logging)
            assert False  # Unreachable.




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import datetime

from freezegun import freeze_time
import pytest

from dob.facts.add_fact import add_fact


class TestAddFact(object):
    """Unit test related to starting a new fact."""

    @freeze_time('2015-12-25 18:00')
    @pytest.mark.parametrize(
        ('raw_fact', 'time_hint', 'expectation'),
        [
            # Use clock-to-clock format, the date inferred from now; with actegory.
            ('13:00 to 16:30: foo@bar', 'verify_both', {
                'activity': 'foo',
                'category': 'bar',
                'start': datetime.datetime(2015, 12, 25, 13, 0, 0),
                'end': datetime.datetime(2015, 12, 25, 16, 30, 0),
                'tags': [],
            }),
            # Use datetime-to-datetime format, with actegory.
            ('2015-12-12 13:00 to 2015-12-12 16:30: foo@bar', 'verify_both', {
                'activity': 'foo',
                'category': 'bar',
                'start': datetime.datetime(2015, 12, 12, 13, 0, 0),
                'end': datetime.datetime(2015, 12, 12, 16, 30, 0),
                'tags': [],
            }),
            # The end date is inferred from start date.
            ('2015-12-12 13:00 - 18:00 foo@bar', 'verify_both', {
                'activity': 'foo',
                'category': 'bar',
                'start': datetime.datetime(2015, 12, 12, 13, 0, 0),
                'end': datetime.datetime(2015, 12, 12, 18, 00, 0),
                'tags': [],
            }),
            # actegory spanning day (straddles) midnight) and spanning multiple days.
            ('2015-12-12 13:00 - 2015-12-25 18:00 foo@bar', 'verify_both', {
                'activity': 'foo',
                'category': 'bar',
                'start': datetime.datetime(2015, 12, 12, 13, 0, 0),
                'end': datetime.datetime(2015, 12, 25, 18, 00, 0),
                'tags': [],
            }),
            # Create open/ongoing/un-ended fact.
            ('2015-12-12 13:00 foo@bar', 'verify_start', {
                'activity': 'foo',
                'category': 'bar',
                'start': datetime.datetime(2015, 12, 12, 13, 0, 0),
                'end': None,
                'tags': [],
            }),
            # Create ongoing fact starting at right now.
            ('foo@bar', 'verify_none', {
                'activity': 'foo',
                'category': 'bar',
                'start': datetime.datetime(2015, 12, 25, 18, 0, 0),
                'end': None,
                'tags': [],
            }),
            # Tags.
            (
                '2015-12-12 13:00 foo@bar: #precious #hashish, i like ike',
                'verify_start',
                {
                    'activity': 'foo',
                    'category': 'bar',
                    'start': datetime.datetime(2015, 12, 12, 13, 0, 0),
                    'end': None,
                    'tags': ['precious', 'hashish'],
                    'description': 'i like ike',
                },
            ),
            # Multiple Tags are identified by a clean leading delimiter character.
            (
                '2015-12-12 13:00 foo@bar, #just walk away "#not a tag", blah',
                'verify_start',
                {
                    'activity': 'foo',
                    'category': 'bar',
                    'start': datetime.datetime(2015, 12, 12, 13, 0, 0),
                    'end': None,
                    'tags': ['just walk away "#not a tag"'],
                    'description': 'blah',
                },
            ),
            # Alternative tag delimiter; and quotes are just consumed as part of tag.
            (
                '2015-12-12 13:00 foo@bar, #just walk away @"totes a tag", blah',
                'verify_start',
                {
                    'activity': 'foo',
                    'category': 'bar',
                    'start': datetime.datetime(2015, 12, 12, 13, 0, 0),
                    'end': None,
                    'tags': ['just walk away', '"totes a tag"'],
                    'description': 'blah',
                },
            ),
            # Test '#' in description, elsewhere, after command, etc.
            (
                '2015-12-12 13:00 baz@bat",'
                ' #tag1, #tag2 tags cannot come #too late, aha!'
                ' Time is also ignored at end: 12:59',
                'verify_start',
                {
                    'activity': 'baz',
                    'category': 'bat"',
                    'start': datetime.datetime(2015, 12, 12, 13, 0, 0),
                    'end': None,
                    'tags': ['tag1'],
                    'description': '#tag2 tags cannot come #too late, aha!'
                                   ' Time is also ignored at end: 12:59',
                },
            ),
        ],
    )
    def test_add_new_fact(
        self,
        controller_with_logging,
        mocker,
        raw_fact,
        time_hint,
        expectation,
    ):
        """
        Test that input validation and assignment of start/end time(s) works as expected.

        To test just this function -- and the parametrize, above -- try:

          workon dob
          cdproject
          py.test --pdb -vv -k test_add_new_fact tests/

        """
        controller = controller_with_logging
        mocker.patch.object(controller.facts, 'save')
        add_fact(controller, raw_fact, time_hint=time_hint, use_carousel=False)
        assert controller.facts.save.called
        args, kwargs = controller.facts.save.call_args
        fact = args[0]
        assert fact.start == expectation['start']
        assert fact.end == expectation['end']
        assert fact.activity.name == expectation['activity']
        assert fact.category.name == expectation['category']
        expecting_tags = ''
        tagnames = list(expectation['tags'])
        if tagnames:
            tagnames.sort()
            expecting_tags = ['#{}'.format(name) for name in tagnames]
            expecting_tags = '{}'.format(' '.join(expecting_tags))
        assert fact.tagnames() == expecting_tags
        expect_description = expectation.get('description', None)
        assert fact.description == expect_description


# ***

class TestStop(object):
    """Unit test concerning the stop command."""

    def test_stop_existing_ongoing_fact(
        self,
        ongoing_fact,
        controller_with_logging,
        mocker,
    ):
        """Make sure stopping an ongoing fact works as intended."""
        mockfact = mocker.MagicMock()
        mockfact.activity.name = 'foo'
        mockfact.category.name = 'bar'
        mocktime = mocker.MagicMock(return_value="%Y-%m-%d %H:%M")
        mockfact.start.strftime = mocktime
        mockfact.end.strftime = mocktime
        current_fact = mocker.MagicMock(return_value=mockfact)
        # While nark still has stop_current_fact, dob replaced stop_fact
        # with add_fact, so it can use all the same CLI magic that the
        # other add-fact commands use. So while we're testing stop-fact
        # here, we're really testing add-fact with a verify-end time-hint.
        controller_with_logging.facts.save = current_fact
        # 2019-12-06: stop_fact was deleted, replaced with add_fact + time_hint.
        add_fact(
            controller_with_logging,
            factoid='',
            time_hint='verify_end',
            use_carousel=False,
        )
        assert controller_with_logging.facts.save.called

    def test_stop_no_existing_ongoing_fact(self, controller_with_logging, capsys):
        """Make sure that stop without actually an ongoing fact leads to an error."""
        with pytest.raises(SystemExit):
            # 2019-12-06: stop_fact was deleted, replaced with add_fact + time_hint.
            add_fact(
                controller_with_logging,
                factoid='',
                time_hint='verify_end',
                use_carousel=False,
            )




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Click Group wrapper tweaks usage for the ``dob add --help`` help."""

from .aliasable_bunchy_plugin import ClickAliasableBunchyPluginGroup
from .bunchy_help import help_header_add_fact

__all__ = (
    'ClickAddFactHelpGroup'
)


class ClickAddFactHelpGroup(ClickAliasableBunchyPluginGroup):

    ADD_FACT_GROUP_NAME = 'add'

    ADD_FACT_GROUP_HELP = 'add --help'

    # SYNC_ME:
    #   dob.run.command(name) ↔ ClickAddFactHelpGroup.ADD_FACT_SKIP_ALIASED=[name,]
    ADD_FACT_SKIP_ALIASED = ['on', 'until', 'next']
    # Here the main names = ['now', 'to', 'after'].
    # FIXME/2019-11-21: (lb): Maybe remove these aliased Add Fact cmds?
    #                        (And be more confident about chosen names.)

    def __init__(self, *args, **kwargs):
        super(ClickAddFactHelpGroup, self).__init__(*args, **kwargs)

    def help_usage_command_path_postfix(self, ctx):
        commands = []
        bunchy_group = ctx.find_root().command.group_bunchies[help_header_add_fact]
        for cmdname in bunchy_group.keys():
            if cmdname in [
                ClickAddFactHelpGroup.ADD_FACT_GROUP_NAME,
                ClickAddFactHelpGroup.ADD_FACT_GROUP_HELP,
                *ClickAddFactHelpGroup.ADD_FACT_SKIP_ALIASED,
            ]:
                continue
            # MEH: (lb): We could weed out hidden commands, too. Seems like work.
            commands.append(cmdname)
        return '|'.join(commands)

    def format_help_text(self, ctx, formatter):
        # (lb): Achoo! Undent (-1, yo!) before click indents (back to 0, we want).
        formatter.dedent()
        super(ClickAddFactHelpGroup, self).format_help_text(ctx, formatter)
        formatter.indent()

